import logging

from apstools.devices.shutters import ApsPssShutterWithStatus as Shutter

from .._iconfig import load_config
from .device import make_device

log = logging.getLogger(__name__)


def load_shutters(config=None):
    if config is None:
        config = load_config()
    # Guard to make sure there's at least one shutter configuration
    if "shutter" not in config.keys():
        return []
    # Load the shutter configurations into devices
    prefix = config["shutter"]["prefix"]
    devices = []
    for name, d in config["shutter"].items():
        if name == "prefix":
            continue
        # Calculate suitable PV values
        hutch = d["hutch"]
        acronym = "FES" if hutch == "A" else f"S{hutch}S"
        devices.append(
            make_device(
                Shutter,
                prefix=f"{prefix}:{acronym}",
                open_pv=f"{prefix}:{acronym}_OPEN_EPICS.VAL",
                close_pv=f"{prefix}:{acronym}_CLOSE_EPICS.VAL",
                state_pv=f"{prefix}:{hutch}_BEAM_PRESENT",
                name=name,
                labels={"shutters"},
            )
        )
    return devices


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



"""Holds ion chamber detector descriptions and assignments to EPICS PVs."""

import logging
import math
import time
import warnings
from collections import OrderedDict
from numbers import Number
from typing import Dict, Generator, Optional

import numpy as np
from apstools.devices.srs570_preamplifier import (
    SRS570_PreAmplifier,
    calculate_settle_time,
)
from ophyd import Component as Cpt
from ophyd import Device, EpicsSignal, EpicsSignalRO
from ophyd import FormattedComponent as FCpt
from ophyd import Kind, Signal, flyers, status
from ophyd.mca import EpicsMCARecord
from ophyd.ophydobj import OphydObject
from ophyd.signal import DerivedSignal, InternalSignal
from ophyd.status import SubscriptionStatus
from pcdsdevices.signal import MultiDerivedSignal, MultiDerivedSignalRO
from pcdsdevices.type_hints import OphydDataType, SignalToValue

from .. import exceptions
from .._iconfig import load_config
from .device import await_for_connection, make_device, resolve_device_names
from .labjack import AnalogInput
from .scaler_triggered import ScalerSignalRO, ScalerTriggered

log = logging.getLogger(__name__)


__all__ = ["IonChamber", "load_ion_chambers"]


class VoltageSignal(DerivedSignal):
    """Calculate the voltage at the output of the pre-amp."""

    def inverse(self, value):
        """Calculate the voltage given a scaler count."""
        clock_ticks = self.parent.clock_ticks.get()
        if clock_ticks == 0:
            return 0
        clock_frequency = self.parent.frequency.get()
        if clock_frequency == 0:
            return 0
        # Convert counts to volt-seconds
        volt_seconds = value / self.parent.counts_per_volt_second
        # Convert volt-seconds to average voltage
        clock_frequency = self.parent.frequency.get()
        seconds = clock_ticks / clock_frequency
        volts = volt_seconds / seconds
        return volts


class CurrentSignal(DerivedSignal):
    """Calculate the current in amps at the input of the pre-amp."""

    def preamp(self):
        """Find which parent in the device hierarchy has the preamp."""
        device = self
        while device is not None:
            if isinstance(getattr(device, "preamp", None), IonChamberPreAmplifier):
                return device.preamp
            else:
                # Move up a step in the hierarchy
                device = device.parent
        # If we get here, there's no pre-amp
        raise AttributeError(f"No ancestor of {self} has a pre-amp.")

    def inverse(self, value):
        """Calculate the current given a output voltage."""
        volts = value
        preamp = self.preamp()
        try:
            gain = preamp.gain.get()
            offset_current = preamp.offset_current.get()
        except TimeoutError:
            msg = (
                "Could not read inverse signals: "
                f"{preamp.gain}, {preamp.offset_current}"
            )
            log.debug(msg)
        else:
            return volts / gain - offset_current


class Voltmeter(AnalogInput):
    amps: OphydObject = Cpt(CurrentSignal, derived_from="volts", kind="normal")
    # Rename ``final_value`` to ``volts``
    final_value = None
    volts = Cpt(EpicsSignal, ".VAL", kind="normal")


class GainDerivedSignal(MultiDerivedSignal):
    """A gain level signal that incorporates dynamic settling time."""

    def set(
        self,
        value: OphydDataType,
        *,
        timeout: Optional[float] = None,
        settle_time: Optional[float] = "auto",
    ):
        # Calculate an auto settling time
        if settle_time == "auto":
            # Determine the new values that will be set
            to_write = self.calculate_on_put(mds=self, value=value) or {}
            # Calculate the correct settling time
            settle_time_ = calculate_settle_time(
                gain_value=to_write[self.parent.sensitivity_value],
                gain_unit=to_write[self.parent.sensitivity_unit],
                gain_mode=self.parent.gain_mode.get(),
            )
        else:
            settle_time_ = settle_time
        # Call the actual set method to move the gain
        return super().set(value, timeout=timeout, settle_time=settle_time_)


class IonChamberPreAmplifier(SRS570_PreAmplifier):
    """An SRS-570 pre-amplifier driven by an ion chamber.

        Has extra signals for walking up and down the sensitivity
        range. *gain_level* is corresponds to the inverse of the
        combination of *sensitivity_value* and *sensitivity_unit*. Setting
        *gain_level* to 0 sets *sensitivity_value* and *sensitivity_unit*
        to "1 mA/V".

    By setting the *gain_level* signal, the offset is
        also set to be 10% of the sensitivity.

    """

    values = ["1", "2", "5", "10", "20", "50", "100", "200", "500"]
    units = ["pA/V", "nA/V", "uA/V", "mA/V"]
    offset_units = [s.split("/")[0] for s in units]
    offset_difference = -3  # How many levels higher should the offset be
    current_multipliers = {
        0: 1e-12,  # pA
        1: 1e-9,  # nA
        2: 1e-6,  # µA
        3: 1e-3,  # mA
        "pA": 1e-12,
        "nA": 1e-9,
        "uA": 1e-6,
        "µA": 1e-6,
        "mA": 1e-3,
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Reset the gain name, apstools converts "preamp_gain" to "preamp"
        self.gain.name += "_gain"
        # Subscriptions for updating the sensitivity text
        self.sensitivity_value.subscribe(self.update_sensitivity_text, run=False)
        self.sensitivity_unit.subscribe(self.update_sensitivity_text, run=True)

    def cb_gain(self, *args, **kwargs):
        """
        Called when sensitivity changes (EPICS CA monitor event).
        """
        gain = self.computed_gain()
        self.gain.put(gain, internal=True)
        self.gain_db.put(10 * math.log10(gain), internal=True)

    def computed_gain(self):
        """
        Amplifier gain (V/A), as floating-point number.
        """
        # Convert the sensitivity to a proper number
        val_idx = int(self.sensitivity_value.get(as_string=False))
        val = float(self.values[val_idx])
        # Determine multiplier based on the gain unit
        amps = [
            1e-12,  # pA
            1e-9,  # nA
            1e-6,  # μA
            1e-3,  # mA
        ]
        unit_idx = int(self.sensitivity_unit.get(as_string=False))
        multiplier = amps[unit_idx]
        inverse_gain = val * multiplier
        return 1 / inverse_gain

    def update_sensitivity_text(self, *args, obj: OphydObject, **kwargs):
        val = self.values[int(self.sensitivity_value.get(as_string=False))]
        unit = self.units[int(self.sensitivity_unit.get(as_string=False))]
        text = f"{val} {unit}"
        self.sensitivity_text.put(text, internal=True)

    def _level_to_value(self, level):
        return level % len(self.values)

    def _level_to_unit(self, level):
        return self.units[int(level / len(self.values))]

    def _get_gain_level(self, mds: MultiDerivedSignal, items: SignalToValue) -> int:
        """Given a sensitivity value and unit, transform to the desired level."""
        value = self.values.index(items[self.sensitivity_value])
        unit = self.units.index(items[self.sensitivity_unit])
        # Determine sensitivity level
        new_level = value + unit * len(self.values)
        # Convert to gain by inverting
        new_level = 27 - new_level
        log.debug(
            f"Getting sensitivity level {self.name}: {value} {unit} -> {new_level}"
        )
        return new_level

    def _put_gain_level(
        self, mds: MultiDerivedSignal, value: OphydDataType
    ) -> SignalToValue:
        "Given a gain level, transform to the desired sensitivity value and unit."
        # Determine new values
        new_level = 27 - value
        new_offset = max(new_level + self.offset_difference, 0)
        # Check for out of bounds
        lmin, lmax = (0, 27)
        msg = (
            f"Cannot set {self.name} outside range ({lmin}, {lmax}), received"
            f" {new_level}."
        )
        if new_level < lmin:
            raise exceptions.GainOverflow(msg)
        elif new_level > lmax:
            raise exceptions.GainOverflow(msg)
        # Return calculated gain and offset
        offset_value = self.values[self._level_to_value(new_offset)]
        offset_unit = self._level_to_unit(new_offset).split("/")[0]
        result = OrderedDict()
        result.update({self.sensitivity_unit: self._level_to_unit(new_level)})
        result.update({self.sensitivity_value: self._level_to_value(new_level)})
        result.update({self.offset_value: offset_value})
        result.update({self.offset_unit: offset_unit})
        # result[self.set_all] = 1
        return result

    def _get_offset_current(
        self, *, mds: MultiDerivedSignal, items: SignalToValue
    ) -> float:
        """Calculate the current in amps added to the signal before amplification."""
        if items[self.offset_on] in ["OFF", "0", 0]:
            return 0
        # Calculate offset current
        val = items[self.offset_value]
        sign = items[self.offset_sign]
        unit = items[self.offset_unit]
        try:
            val = float(f"{sign}{val}")
            multiplier = self.current_multipliers[unit]
            current = val * multiplier
        except ValueError:
            return 0
        return current

    gain_level = Cpt(
        GainDerivedSignal,
        attrs=[
            "sensitivity_value",
            "sensitivity_unit",
            "offset_value",
            "offset_unit",
            "set_all",
        ],
        calculate_on_get=_get_gain_level,
        calculate_on_put=_put_gain_level,
        kind=Kind.omitted,
    )
    offset_current = Cpt(
        MultiDerivedSignalRO,
        attrs=["offset_value", "offset_unit", "offset_on", "offset_sign"],
        calculate_on_get=_get_offset_current,
        kind=Kind.config,
    )

    # A text description of the what the current sensitivity settings are
    sensitivity_text = Cpt(
        InternalSignal,
        kind=Kind.config,
    )
    # Gain, but measured in various forms
    gain = Cpt(InternalSignal, name="gainerificf", kind="normal", value=1)
    gain_db = Cpt(InternalSignal, kind=Kind.config, value=0)


class IonChamber(ScalerTriggered, Device, flyers.FlyerInterface):
    """An ion chamber at a spectroscopy beamline.

    Also includes the pre-amplifier as ``.pre_amp``.

    This class also implements the bluesky/ophyd flyer
    interface. During *kickoff()*, previous data are erased and
    acquisition is started as a multi-channel scaler. It also watches
    for changes in the number of data collected and collects
    timestamps each time a new datum is captured. *complete()* stops
    acquisition. These timestamps, along with the measured data, are
    generated during *collect()*.

    Parameters
    ==========
    prefix
      The PV prefix of the overall scaler.
    ch_num
      The number (1-index) of the channel on the scaler. 1 is the
      timer, so your channel number should start at 2.
    name
      The bluesky-compatible name for this device.
    preamp_prefix
      The process variable prefix to the pre-amp that controls this
      ion chamber (e.g. "25idc:SR01").
    scaler_prefix
      The process variable prefix for the scaler that measures this
      ion chamber.
    voltmeter_prefix
      The process variable prefix for the voltmeter. This might be an
      analog input for a labjack, in which case try something like
      "LabJackT7_1:Ai0".

    Attributes
    ==========
    ch_num
      The channel number on the scaler, starting at 2 (1 is the timer).
    count
      The trigger to count scaler pulses.
    counts
      The counts coming from the scaler without any correction.
    volts
      The volts produced by the pre-amp, calculated from scaler
      counts.
    amps
      The current produced by the ion chamber, calculated from scaler
      counts and preamp settings.
    exposure_time
      Positioner for setting the count time on the scaler.
    preamp
      The SR570 pre-amplifier driving the signal.

    """

    stream_name: str = "primary"
    ch_num: int = 0
    ch_char: str
    start_timestamp: float = None
    count: OphydObject = FCpt(
        EpicsSignal, "{scaler_prefix}scaler1.CNT", trigger_value=1, kind=Kind.omitted
    )
    description: OphydObject = FCpt(
        EpicsSignal, "{scaler_prefix}scaler1.NM{ch_num}", kind=Kind.config
    )
    # Signal chain devices
    preamp = FCpt(IonChamberPreAmplifier, "{preamp_prefix}")
    voltmeter = FCpt(Voltmeter, "{voltmeter_prefix}", kind=Kind.hinted)
    # Measurement signals
    volts: OphydObject = Cpt(VoltageSignal, derived_from="counts", kind=Kind.normal)
    amps: OphydObject = Cpt(CurrentSignal, derived_from="volts", kind=Kind.hinted)
    counts: OphydObject = FCpt(
        EpicsSignalRO,
        "{scaler_prefix}scaler1.S{ch_num}",
        kind=Kind.normal,
        auto_monitor=False,
    )
    gate: OphydObject = FCpt(
        EpicsSignal,
        "{scaler_prefix}scaler1.G{ch_num}",
        kind=Kind.config,
    )
    preset_count: OphydObject = FCpt(
        EpicsSignal, "{scaler_prefix}scaler1.PR{ch_num}", kind=Kind.config
    )
    frequency: OphydObject = FCpt(
        EpicsSignal,
        "{scaler_prefix}scaler1.FREQ",
        kind=Kind.config,
    )
    clock_ticks: OphydObject = FCpt(
        EpicsSignalRO,
        "{scaler_prefix}scaler1.S1",
        kind=Kind.normal,
    )
    # Old Scaler mode support
    offset: OphydObject = FCpt(
        ScalerSignalRO, "{scaler_prefix}scaler1_{offset_suffix}", kind=Kind.config
    )
    net_counts: OphydObject = FCpt(
        ScalerSignalRO, "{scaler_prefix}scaler1_netA.{ch_char}", kind=Kind.hinted
    )
    exposure_time: OphydObject = FCpt(
        EpicsSignal, "{scaler_prefix}scaler1.TP", kind=Kind.normal
    )
    auto_count: OphydObject = FCpt(
        EpicsSignal, "{scaler_prefix}scaler1.CONT", kind=Kind.omitted
    )
    record_dark_current: OphydObject = FCpt(
        EpicsSignal, "{scaler_prefix}scaler1_offset_start.PROC", kind=Kind.omitted
    )
    record_dark_time: OphydObject = FCpt(
        EpicsSignal, "{scaler_prefix}scaler1_offset_time.VAL", kind=Kind.config
    )
    # Multi-channel scaler support
    start_all: OphydObject = FCpt(
        EpicsSignal, "{scaler_prefix}StartAll", kind=Kind.omitted
    )
    stop_all: OphydObject = FCpt(
        EpicsSignal, "{scaler_prefix}StopAll", kind=Kind.omitted
    )
    erase_all: OphydObject = FCpt(
        EpicsSignal, "{scaler_prefix}EraseAll", kind=Kind.omitted
    )
    erase_start: OphydObject = FCpt(
        EpicsSignal,
        "{scaler_prefix}EraseStart",
        kind=Kind.omitted,
    )
    acquiring: OphydObject = FCpt(
        EpicsSignal, "{scaler_prefix}Acquiring", kind=Kind.omitted
    )
    channel_advance_source: OphydObject = FCpt(
        EpicsSignal,
        "{scaler_prefix}ChannelAdvance",
        kind=Kind.config,
    )
    num_channels_to_use: OphydObject = FCpt(
        EpicsSignal,
        "{scaler_prefix}NuseAll",
        kind=Kind.config,
    )
    max_channels: OphydObject = FCpt(
        EpicsSignal, "{scaler_prefix}MaxChannels", kind=Kind.config
    )
    current_channel: OphydObject = FCpt(
        EpicsSignal,
        "{scaler_prefix}CurrentChannel",
        kind=Kind.normal,
    )
    channel_one_source: OphydObject = FCpt(
        EpicsSignal, "{scaler_prefix}Channel1Source", kind=Kind.config
    )
    count_on_start: OphydObject = FCpt(
        EpicsSignal, "{scaler_prefix}CountOnStart", kind=Kind.config
    )
    mca: OphydObject = FCpt(
        EpicsMCARecord, "{scaler_prefix}mca{ch_num}", kind=Kind.omitted
    )
    mca_times: OphydObject = FCpt(
        EpicsMCARecord, "{scaler_prefix}mca1", kind=Kind.omitted
    )

    # Virtual signals to handle fly-scanning
    timestamps: list = []
    num_bins = Cpt(Signal)

    _default_read_attrs = [
        "counts",
        "volts",
        "exposure_time",
        "net_counts",
        "voltmeter",
    ]

    def __init__(
        self,
        prefix: str,
        ch_num: int,
        name: str,
        preamp_prefix: str = None,
        scaler_prefix: str = None,
        voltmeter_prefix: str = None,
        counts_per_volt_second: Number = 1,
        *args,
        **kwargs,
    ):
        self.counts_per_volt_second = counts_per_volt_second
        # Set up the channel number for this scaler channel
        if ch_num < 1:
            raise ValueError(f"Scaler channels must be greater than 0: {ch_num}")
        self.ch_num = ch_num
        self.ch_char = self.num_to_char(ch_num)
        # Determine which prefix to use for the scaler
        if scaler_prefix is None:
            scaler_prefix = prefix
        self.scaler_prefix = scaler_prefix
        # Save an epics path to the preamp
        if preamp_prefix is None:
            preamp_prefix = prefix
        self.preamp_prefix = preamp_prefix
        # Save an epics path for the voltmeter (nominally a labjack)
        self.voltmeter_prefix = voltmeter_prefix
        # Determine the offset PV since it follows weird numbering conventions
        calc_num = int((self.ch_num - 1) / 4)
        calc_char = self.num_to_char(((self.ch_num - 1) % 4) + 1)
        self.offset_suffix = f"offset{calc_num}.{calc_char}"
        # Initialize all the other Device stuff
        super().__init__(prefix=prefix, name=name, *args, **kwargs)
        # Set signal values to stage
        self.stage_sigs[self.auto_count] = 0

    @property
    def default_time_signal(self):
        """The signal to use for setting exposure time when no other signal is
        provided.

        """
        return self.exposure_time

    def num_to_char(self, num):
        char = chr(64 + num)
        return char

    def change_sensitivity(self, step: int) -> status.StatusBase:
        """Change the gain on the pre-amp by the given number of steps.

        Parameters
        ==========
        step
          How many levels to change the sensitivity. Positive numbers
          increase the gain, negative numbers decrease the gain.

        Returns
        =======
        status.StatusBase
          The status that will be marked complete once the sensitivity
          is changed.

        """
        new_sens_level = self.sensitivity.sens_level.readback.get() + step
        try:
            status = self.sensitivity.sens_level.set(new_sens_level)
        except ValueError as e:
            raise exceptions.GainOverflow(f"{self.name} -> {e}")
        return status

    def increase_gain(self) -> status.StatusBase:
        """Increase the gain (descrease the sensitivity) of the ion chamber's
        pre-amp.

        Returns
        =======
        status.StatusBase
          Ophyd status object for the value and gain of the
          sensitivity in the pre-amp.

        """
        return self.change_sensitivity(-1)

    def decrease_gain(self) -> status.StatusBase:
        """Decrease the gain (increase the sensitivity) of the ion chamber's
        pre-amp.

        Returns
        =======
        status.StatusBase
          Ophyd status object for the value and gain of the
          sensitivity in the pre-amp.

        """
        return self.change_sensitivity(1)

    def record_timestamp(self, *, old_value, value, timestamp, **kwargs):
        self.timestamps.append(timestamp)

    def kickoff(self) -> status.StatusBase:
        def check_acquiring(*, old_value, value, **kwargs):
            is_acquiring = bool(value)
            if is_acquiring:
                self.start_timestamp = time.time()
            return is_acquiring

        self.start_timestamp = None
        # Set some configuration PVs on the MCS
        self.count_on_start.set(1).wait()
        # Start acquiring data
        self.erase_start.set(1).wait()
        # Wait for the "Acquiring" to start
        status = SubscriptionStatus(self.acquiring, check_acquiring)
        # Watch for new data being collected so we can save timestamps
        self.timestamps = []
        self.current_channel.subscribe(self.record_timestamp)
        return status

    def complete(self) -> status.StatusBase:
        status = self.stop_all.set(1, settle_time=0.05)
        return status

    def collect(self) -> Generator[Dict, None, None]:
        net_counts_name = self.net_counts.name
        # Use the scaler's clock counter to calculate timestamps
        times = self.mca_times.spectrum.get()
        times = np.divide(times, self.frequency.get(), casting="safe")
        times = np.cumsum(times)
        pso_timestamps = times + self.start_timestamp
        # Retrieve data, except for first point (during taxiing)
        data = self.mca.spectrum.get()[1:]
        # Convert timestamps from PSO pulses to pixels
        pixel_timestamps = (pso_timestamps[1:] + pso_timestamps[:-1]) / 2
        # Create data events
        for ts, value in zip(pixel_timestamps, data):
            yield {
                "data": {net_counts_name: value},
                "timestamps": {net_counts_name: ts},
                "time": ts,
            }

    def describe_collect(self) -> Dict[str, Dict]:
        """Describe details for the flyer collect() method"""
        desc = OrderedDict()
        desc.update(self.net_counts.describe())
        return {self.name: desc}


async def make_ion_chamber_device(
    prefix: str, ch_num: int, name: str, preamp_prefix: str
):
    ic = IonChamber(
        prefix=prefix,
        ch_num=ch_num,
        name=name,
        preamp_prefix=preamp_prefix,
        labels={"ion_chambers", "detectors"},
    )
    try:
        await await_for_connection(ic)
    except TimeoutError as exc:
        log.warning(
            f"Could not connect to ion chamber: {name} ({prefix}, {preamp_prefix})"
        )
    else:
        log.info(f"Created ion chamber: {name} ({prefix}, {preamp_prefix})")
        return ic


def load_ion_chamber(
    preamp_prefix: str,
    scaler_prefix: str,
    voltmeter_prefix: str,
    ch_num: int,
    name: str,
):
    """Create an IonChamber ophyd device.

    When autoloading the ion chambers from the config file, this
    function will parse the config file arguments and convert them
    into the forms needed for the IonChamber device itself.

    """
    # Determine ion_chamber configuration
    preamp_prefix = f"{preamp_prefix}:SR{ch_num:02}:"
    desc_pv = f"{scaler_prefix}:scaler1.NM{ch_num}"
    # Determine which labjack channel is measuring the voltmeter
    ic_idx = ch_num - 2
    # 5 pre-amps per labjack
    lj_num = int(ic_idx / 5)
    lj_chan = ic_idx % 5
    # Create the ion chamber device
    ion_chamber = make_device(
        IonChamber,
        prefix=scaler_prefix,
        ch_num=ch_num,
        name=name,
        preamp_prefix=preamp_prefix,
        voltmeter_prefix=f"{voltmeter_prefix}{lj_num}:Ai{lj_chan}",
        labels={"ion_chambers", "detectors"},
    )
    return ion_chamber


async def load_ion_chambers(config=None):
    """Load ion chambers based on configuration files' ``[ion_chamber]``
    sections.

    The name for each ion chamber is retrieved from the scaler
    channel's .DESC field.

    """
    # Load IOC prefixes from the config file
    if config is None:
        config = load_config()
    if "ion_chamber" not in config.keys():
        warnings.warn("Ion chambers not configured.")
        return []
    # Generate the configuration dictionary for all the ion chambers
    ic_defns = []
    for section_name, section in config["ion_chamber"].items():
        channels = zip(
            section["scaler_channels"],
            section["preamp_channels"],
            section["voltmeter_channels"],
        )
        for scaler_ch, preamp_ch, voltmeter_ch in channels:
            voltmeter_prefix = f"{section['voltmeter_prefix']}Ai{voltmeter_ch}"
            preamp_prefix = f"{section['preamp_prefix']}{preamp_ch:02}:"
            scaler_prefix = section["scaler_prefix"]
            desc_pv = f"{scaler_prefix}scaler1.NM{scaler_ch}"
            ic_defns.append(
                {
                    "section": section_name,
                    "scaler_prefix": section["scaler_prefix"],
                    "ch_num": scaler_ch,
                    "voltmeter_prefix": voltmeter_prefix,
                    "preamp_prefix": preamp_prefix,
                    "desc_pv": desc_pv,
                    "counts_per_volt_second": section["counts_per_volt_second"],
                }
            )
    # Resolve the scaler channels into ion chamber names
    await resolve_device_names(ic_defns)
    # Loop through the sections and create ion chambers
    devices = []
    missing_channels = []
    unnamed_channels = []
    for defn in ic_defns:
        if defn["name"] == "":
            unnamed_channels.append(defn["desc_pv"])
        elif defn["name"] is None:
            missing_channels.append(defn["desc_pv"])
        else:
            # Create the ion chamber device
            devices.append(
                make_device(
                    IonChamber,
                    prefix=defn["scaler_prefix"],
                    ch_num=defn["ch_num"],
                    name=defn["name"],
                    preamp_prefix=defn["preamp_prefix"],
                    voltmeter_prefix=defn["voltmeter_prefix"],
                    labels={"ion_chambers", defn["section"], "detectors"},
                    counts_per_volt_second=defn["counts_per_volt_second"],
                )
            )
    # Notify of any missing ion chambers
    if len(missing_channels) > 0:
        msg = "Skipping unavailable ion chambers: "
        msg += ", ".join([prefix for prefix in missing_channels])
        warnings.warn(msg)
        log.warning(msg)
    if len(unnamed_channels) > 0:
        msg = "Skipping unnamed ion chambers: "
        msg += ", ".join([prefix for prefix in unnamed_channels])
        warnings.warn(msg)
        log.warning(msg)
    return devices


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from bluesky import plan_stubs as bps

from ..instrument.instrument_registry import registry
from .shutters import close_shutters, open_shutters


def record_dark_current(ion_chambers, shutters, time):
    """Record the dark current on the ion chambers.

    - Close shutters
    - Record ion chamber dark current
    - Open shutters

    Parameters
    ==========
    ion_chambers
      Ion chamber devices or names.
    shutters
      Shutter devices or names.

    """
    yield from close_shutters(shutters)
    # Measure the dark current
    ion_chambers = registry.findall(ion_chambers)
    # This is a big hack, we need to come back and just accept the current integration time
    old_times = [ic.exposure_time.get() for ic in ion_chambers]
    time_args = [obj for ic in ion_chambers for obj in (ic.record_dark_time, time)]
    yield from bps.mv(*time_args)
    mv_args = [obj for ic in ion_chambers for obj in (ic.record_dark_current, 1)]
    triggers = [ic.record_dark_current for ic in ion_chambers]
    yield from bps.mv(*mv_args)
    yield from bps.sleep(time)
    time_args = [
        obj for (ic, t) in zip(ion_chambers, old_times) for obj in (ic.exposure_time, t)
    ]
    yield from bps.mv(*time_args)
    # Open shutters again
    yield from open_shutters(shutters)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import asyncio
import logging
import warnings
from typing import Mapping

from ophyd import sim

from .._iconfig import load_config
from .aerotech import load_aerotech_stages
from .aps import load_aps
from .area_detector import load_area_detectors
from .beamline_manager import load_beamline_manager
from .camera import load_cameras
from .dxp import load_dxp_detectors
from .energy_positioner import load_energy_positioner
from .heater import load_heaters
from .instrument_registry import InstrumentRegistry
from .instrument_registry import registry as default_registry
from .ion_chamber import load_ion_chambers
from .lerix import load_lerix_spectrometers
from .mirrors import load_mirrors
from .motor import HavenMotor, load_motors
from .power_supply import load_power_supplies
from .robot import load_robots
from .shutter import load_shutters
from .slits import load_slits
from .stage import load_stages
from .table import load_tables
from .xia_pfcu import load_xia_pfcu4s
from .xspress import load_xspress_detectors

__all__ = ["load_instrument"]

log = logging.getLogger(__name__)


async def load_instrument(
    registry: InstrumentRegistry = default_registry,
    config: Mapping = None,
    wait_for_connection: bool = True,
    timeout: int = 5,
    return_devices: bool = False,
):
    """Load the beamline instrumentation.

    This function will reach out and query various IOCs for motor
    information based on the information in *config* (see
    ``iconfig_default.toml`` for examples). Based on the
    configuration, it will create Ophyd devices and register them with
    *registry*.

    Parameters
    ==========
    registry:
      The registry into which the ophyd devices will be placed.
    config:
      The beamline configuration read in from TOML files. Mostly
      useful for testing.
    wait_for_connection
      If true, only connected devices will be kept.
    timeout
      How long to wait for if *wait_for_connection* is true.
    return_devices
      If true, return the newly loaded devices when complete.

    """
    # Clear out any existing registry entries
    if registry is not None:
        registry.clear()
    # Load the configuration
    if config is None:
        config = load_config()
    # Asynchronous loading of devices
    results = await asyncio.gather(
        load_ion_chambers(config=config),
        # Load the motor devices last so that we can check for
        # existing motors in the registry
        load_motors(config=config),
    )
    # Flatten async devices
    devices = [d for devs in results for d in devs]
    # Synchronous loading of devices
    devices.extend(
        [
            *load_aerotech_stages(config=config),
            load_aps(config=config),
            *load_area_detectors(config=config),
            load_beamline_manager(config=config),
            *load_cameras(config=config),
            *load_dxp_detectors(config=config),
            load_energy_positioner(config=config),
            *load_heaters(config=config),
            *load_lerix_spectrometers(config=config),
            *load_power_supplies(config=config),
            *load_robots(config=config),
            *load_shutters(config=config),
            *load_slits(config=config),
            *load_stages(config=config),
            *load_tables(config=config),
            *load_xia_pfcu4s(config=config),
            *load_xspress_detectors(config=config),
            *load_mirrors(config=config),
        ]
    )
    # Also import some simulated devices for testing
    # devices.extend(load_simulated_devices(config=config))
    # Filter out devices that couldn't be reached
    devices = [d for d in devices if d is not None]
    # Put the devices into the registry
    if not getattr(registry, "auto_register", True):
        [registry.register(device) for device in devices]
    # Only keep connected devices
    disconnected = []
    if wait_for_connection and registry is not None:
        disconnected = registry.pop_disconnected(timeout=timeout)
        devices = [dev for dev in devices if dev not in disconnected]
        if len(disconnected) > 0:
            msg = "Removed disconnected devices: "
            msg += ", ".join(dev.name for dev in disconnected)
            warnings.warn(msg)
            log.warning(msg)
    # Return the final list
    if return_devices:
        return devices


def load_simulated_devices(config={}):
    # Motors
    FakeMotor = sim.make_fake_device(HavenMotor)
    motor = FakeMotor(name="sim_motor", labels={"motors"})
    # Detectors
    detector = sim.SynGauss(
        name="sim_detector",
        motor=motor,
        motor_field="sim_motor",
        labels={"detectors"},
        center=0,
        Imax=1,
    )
    return (motor, detector)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import qtawesome as qta

from firefly import display


class SlitsMotorDisplay(display.FireflyDisplay):
    def customize_ui(self):
        # Make the tweak buttons use proper arrow icons
        title = self.macros()["TITLE"]
        if "Size" in title:
            forward_icon = qta.icon("fa5s.plus")
            reverse_icon = qta.icon("fa5s.minus")
        elif "Vertical Center" in title:
            forward_icon = qta.icon("fa5s.arrow-up")
            reverse_icon = qta.icon("fa5s.arrow-down")
        else:
            forward_icon = qta.icon("fa5s.arrow-right")
            reverse_icon = qta.icon("fa5s.arrow-left")
        self.ui.tweak_forward_button.setIcon(forward_icon)
        self.ui.tweak_reverse_button.setIcon(reverse_icon)
        for btn in [self.ui.tweak_reverse_button, self.ui.tweak_forward_button]:
            btn.setText("")

    def ui_filename(self):
        return "slits_motor.ui"


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from unittest import mock

import pytest
from ophyd.sim import make_fake_device

from firefly.kb_mirrors import KBMirrorsDisplay
from haven.instrument.mirrors import KBMirrors


@pytest.fixture()
def kb_mirrors(sim_registry):
    """A fake set of slits using the 4-blade setup."""
    FakeMirrors = make_fake_device(KBMirrors)
    kb = FakeMirrors(
        prefix="255idc:KB:",
        name="kb_mirrors",
        horiz_upstream_motor="255idc:m1",
        vert_upstream_motor="255idc:m2",
        horiz_downstream_motor="255idc:m3",
        vert_downstream_motor="255idc:m4",
        labels={"kb_mirrors"},
    )
    return kb


@pytest.fixture()
def kb_bendable_mirrors(sim_registry):
    """A fake set of slits using the 4-blade setup."""
    FakeMirrors = make_fake_device(KBMirrors)
    kb = FakeMirrors(
        prefix="255idc:Long_KB:",
        name="kb_bendable_mirrors",
        horiz_upstream_motor="255idc:m5",
        vert_upstream_motor="255idc:m6",
        horiz_downstream_motor="255idc:m7",
        vert_downstream_motor="255idc:m8",
        horiz_upstream_bender="255idc:m21",
        vert_upstream_bender="255idc:m22",
        horiz_downstream_bender="255idc:m23",
        vert_downstream_bender="255idc:m24",
        labels={"kb_mirrors"},
    )
    return kb


@pytest.fixture()
def display(kb_mirrors, qtbot):
    disp = KBMirrorsDisplay(macros={"DEVICE": kb_mirrors.name})
    qtbot.addWidget(disp)
    return disp


def test_bender_widgets(qtbot, kb_bendable_mirrors):
    display = KBMirrorsDisplay(macros={"DEVICE": kb_bendable_mirrors.name})
    qtbot.addWidget(display)
    # Check that the bender control widgets were enabled
    assert display.ui.horizontal_upstream_display.isEnabled()
    assert display.ui.horizontal_downstream_display.isEnabled()
    assert display.ui.vertical_upstream_display.isEnabled()
    assert display.ui.vertical_downstream_display.isEnabled()


def test_kb_mirrors_caqtdm(display, kb_mirrors):
    display._open_caqtdm_subprocess = mock.MagicMock()
    # Launch the caqtdm display
    display.launch_caqtdm()
    assert display._open_caqtdm_subprocess.called
    cmds = display._open_caqtdm_subprocess.call_args[0][0]
    # Check that the right macros are sent
    # /net/s25data/xorApps/ui/KB_mirrors.ui, macro:
    # P=25idcVME:,PM=25idcVME:,KB=KB,KBH=KB:H,KBV=KB:V,KBHUS=m6,KBHDS=m5,KBVUS=m8,KBVDS=m7,KB1=KBH, KB2=KBV
    macros = [cmds[i + 1] for i in range(len(cmds)) if cmds[i] == "-macro"][0]
    assert "P=255idc:" in macros
    assert "KB=KB" in macros
    assert "KBH=KB:H" in macros
    assert "KBV=KB:V" in macros
    assert "KBHUS=m1" in macros
    assert "KBHDS=m3" in macros
    assert "KBVUS=m2" in macros
    assert "KBVDS=m4" in macros
    assert "KB1=KBH" in macros
    assert "KB2=KBV" in macros
    # Check that the right UI file is being used
    ui_file = cmds[-1]
    assert ui_file.split("/")[-1] == "KB_mirrors.ui"


def test_kb_bendable_mirrors_caqtdm(kb_bendable_mirrors):
    mirrors = kb_bendable_mirrors
    disp = KBMirrorsDisplay(macros={"DEVICE": mirrors.name})
    disp._open_caqtdm_subprocess = mock.MagicMock()
    # Launch the caqtdm display
    disp.launch_caqtdm()
    assert disp._open_caqtdm_subprocess.called
    cmds = disp._open_caqtdm_subprocess.call_args[0][0]
    # Check that the right macros are sent
    # /net/s25data/xorApps/ui/KB_mirrors_and_benders.ui, macro: P=25idcVME:,PM=25idcVME:,KB=LongKB_Cdn,KBH=LongKB_Cdn:H,KBV=LongKB_Cdn:V,KBHUS=m50,KBHDS=m49,KBVUS=m56,KBVDS=m55,KB1=LongKB_CdnH, KB2=LongKB_CdnV,HBUS=m52,HBDS=m51,VBUS=m54,VBDS=m53
    macros = [cmds[i + 1] for i in range(len(cmds)) if cmds[i] == "-macro"][0]
    assert "P=255idc:" in macros
    assert "PM=255idc:" in macros
    assert "KB=Long_KB" in macros
    assert "KBH=Long_KB:H" in macros
    assert "KBV=Long_KB:V" in macros
    assert "KBHUS=m5" in macros
    assert "KBHDS=m7" in macros
    assert "KBVUS=m6" in macros
    assert "KBVDS=m8" in macros
    assert "KB1=Long_KBH" in macros
    assert "KB2=Long_KBV" in macros
    assert "HBUS=m21" in macros
    assert "HBDS=m23" in macros
    assert "VBUS=m22" in macros
    assert "VBDS=m24" in macros
    # Check that the right UI file is being used
    ui_file = cmds[-1]
    assert ui_file.split("/")[-1] == "KB_mirrors_and_benders.ui"



from unittest import mock

import pytest

from firefly.ion_chamber import IonChamberDisplay


@pytest.fixture()
def display(I0, qtbot):
    display = IonChamberDisplay(macros={"IC": I0.name})
    qtbot.addWidget(display)
    display.launch_caqtdm = mock.MagicMock()
    return display


def test_caqtdm_actions(display):
    assert len(display.caqtdm_actions) == 3
    # Check the details for the caQtDM actions
    actions = display.caqtdm_actions
    assert actions[0].text() == "&Scaler caQtDM"
    assert actions[1].text() == "&MCS caQtDM"
    assert actions[2].text() == "&Preamp caQtDM"
    # Check the slots are connected properly


def test_preamp_caqtdm_macros(display):
    # Check that the various caqtdm calls set up the right macros
    display.launch_preamp_caqtdm()
    assert display.launch_caqtdm.called
    assert (
        display.launch_caqtdm.call_args[1]["ui_file"]
        == "/net/s25data/xorApps/epics/synApps_6_2_1/support/ip-GIT/ipApp/op/ui/autoconvert/SR570.ui"
    )
    assert display.launch_caqtdm.call_args[1]["macros"] == {
        "P": "preamp_ioc:",
        "A": "SR04:",
    }


# def test_mcs_caqtdm_macros(display):
#     # Check that the various caqtdm calls set up the right macros
#     display.launch_mcs_caqtdm()
#     assert display.launch_caqtdm.called
#     assert (
#         display.launch_caqtdm.call_args[1]["ui_file"]
#         == "/APSshare/epics/synApps_6_2_1/support/mca-R7-9//mcaApp/op/ui/autoconvert/SIS38XX.ui"
#     )
#     assert display.launch_caqtdm.call_args[1]["macros"] == {"P": "scaler_ioc:"}


@pytest.mark.skip(reason="Causes CI to hang. Not sure why.")
def test_scaler_caqtdm_macros(display):
    # Check that the various caqtdm calls set up the right macros
    display.launch_scaler_caqtdm()
    assert display.launch_caqtdm.called
    assert (
        display.launch_caqtdm.call_args[1]["ui_file"]
        == "/net/s25data/xorApps/ui/scaler32_full_offset.ui"
    )
    assert display.launch_caqtdm.call_args[1]["macros"] == {
        "P": "scaler_ioc:",
        "S": "scaler1",
    }



import logging
import math
import threading
import time
from collections import OrderedDict
from typing import Dict, Generator

import numpy as np
import pint
from apstools.synApps.asyn import AsynRecord
from ophyd import Component as Cpt
from ophyd import FormattedComponent as FCpt
from ophyd import Kind, Signal
from ophyd.status import SubscriptionStatus

from .._iconfig import load_config
from ..exceptions import InvalidScanParameters
from .delay import DG645Delay
from .device import make_device
from .motor import HavenMotor
from .stage import XYStage

log = logging.getLogger(__name__)

ureg = pint.UnitRegistry()


class AerotechFlyer(HavenMotor):
    """Allow an Aerotech stage to fly-scan via the Ophyd FlyerInterface.

    Set *start_position*, *end_position*, and *step_size* in units of
    the motor record (.EGU), and *dwell_time* in seconds. Then the
    remaining components will be calculated accordingly.

    All position or distance components are assumed to be in motor
    record engineering units, unless preceded with "encoder_", in
    which case they are in units of encoder pulses based on the
    encoder resolution.

    The following diagram describes how the various components relate
    to each other. Distances are not to scale::

                 ┌─ encoder_window_start         ┌─ encoder_window_stop
                 │                               │
                 │ |┄┄┄| *step_size*             │
                 │ │   │ encoder_step_size       │
                 │ │   │                         │
      Window:    ├┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┄┤

      Pulses: ┄┄┄┄┄╨───╨───╨───╨───╨───╨───╨───╨┄┄┄┄┄
               │   │ │                       │ │   └─ taxi_end
               │   │ └─ *start_position*     │ └─ pso_end
               │   └─ pso_start              └─ *end position*
               └─ taxi_start

    Parameters
    ==========
    axis
      The label used by the aerotech controller to refer to this
      axis. Examples include "@0" or "X".
    encoder
      The number of the encoder to track when fly-scanning with this
      device.

    Components
    ==========
    start_position
      User-requested center of the first scan pixel.
    end_position
      User-requested center of the last scan pixel. This is not
      guaranteed and may be adjusted to match the encoder resolution
      of the stage.
    step_size
      How much space desired between points. This is not guaranteed
      and may be adjusted to match the encoder resolution of the
      stage.
    dwell_time
      How long to take, in seconds, moving between points.
    slew_speed
      How fast to move the stage. Calculated from the remaining
      components.
    taxi_start
      The starting point for motor movement during flying, accounts
      for needed acceleration of the motor.
    taxi_end
      The target point for motor movement during flying, accounts
      for needed acceleration of the motor.
    pso_start
      The motor position corresponding to the first PSO pulse.
    pso_end
      The motor position corresponding to the last PSO pulse.
    encoder_step_size
      The number of encoder counts for each pixel.
    encoder_window_start
      The start of the window within which PSO pulses may be emitted,
      in encoder counts. Should be slightly wider than the actual PSO
      range.
    encoder_window_end
      The end of the window within which PSO pulses may be emitted,
      in encoder counts. Should be slightly wider than the actual PSO
      range.

    """

    axis: str
    pixel_positions: np.ndarray = None
    # Internal encoder in the Ensemble to track for flying
    encoder: int
    encoder_direction: int = 1
    encoder_window_min: int = -8388607
    encoder_window_max: int = 8388607

    # Calculated fly-scan signals
    pso_start = Cpt(Signal, kind=Kind.config)
    pso_end = Cpt(Signal, kind=Kind.config)
    encoder_step_size = Cpt(Signal, kind=Kind.config)
    encoder_window_start = Cpt(Signal, kind=Kind.config)
    encoder_window_end = Cpt(Signal, kind=Kind.config)
    disable_window = Cpt(Signal, value=False, kind=Kind.config)
    encoder_use_window = Cpt(Signal, value=False, kind=Kind.config)

    # Status signals
    flying_complete = Cpt(Signal, kind=Kind.omitted)
    ready_to_fly = Cpt(Signal, kind=Kind.omitted)

    def __init__(self, *args, axis: str, encoder: int, **kwargs):
        super().__init__(*args, **kwargs)
        # Set up extra calculations for the flyer
        self.encoder_resolution.subscribe(self._update_fly_params)
        self.disable_window.subscribe(self._update_fly_params)
        # Save needed axis/encoder values
        self.axis = axis
        self.encoder = encoder

    def kickoff(self):
        """Start a flyer

        The status object return is marked as done once flying
        has started.

        Returns
        -------
        kickoff_status : StatusBase
            Indicate when flying has started.

        """

        def flight_check(*args, old_value, value, **kwargs) -> bool:
            return not bool(old_value) and bool(value)

        # Status object is complete when flying has started
        self.ready_to_fly.set(False).wait()
        status = SubscriptionStatus(self.ready_to_fly, flight_check)
        # Taxi the motor
        th = threading.Thread(target=self.taxi)
        th.daemon = True
        th.start()
        # Record time of fly start of scan
        self.starttime = time.time()
        self._taxi_thread = th  # Prevents garbage collection
        return status

    def complete(self):
        """Wait for flying to be complete.

        This can either be a question ("are you done yet") or a
        command ("please wrap up") to accommodate flyers that have a
        fixed trajectory (ex. high-speed raster scans) or that are
        passive collectors (ex MAIA or a hardware buffer).

        In either case, the returned status object should indicate when
        the device is actually finished flying.

        Returns
        -------
        complete_status : StatusBase
            Indicate when flying has completed
        """

        # Prepare a callback to check when the motor has stopped moving
        def check_flying(*args, old_value, value, **kwargs) -> bool:
            "Check if flying is complete."
            return bool(value)

        # Status object is complete when flying has started
        self.flying_complete.set(False).wait()
        status = SubscriptionStatus(self.flying_complete, check_flying)
        # Iniate the fly scan
        th = threading.Thread(target=self.fly)
        th.start()
        self._fly_thread = th  # Prevents garbage collection
        return status

    def collect(self) -> Generator[Dict, None, None]:
        """Retrieve data from the flyer as proto-events
        Yields
        ------
        event_data : dict
            Must have the keys {'time', 'timestamps', 'data'}.

        """
        # np array of pixel location
        pixels = self.pixel_positions
        # time of scans start taken at Kickoff
        starttime = self.starttime
        # time of scans at movement stop
        endtime = self.endtime
        # grab necessary for calculation
        accel_time = self.acceleration.get()
        dwell_time = self.flyer_dwell_time.get()
        step_size = self.flyer_step_size()
        slew_speed = step_size / dwell_time
        motor_accel = slew_speed / accel_time
        # Calculate the time it takes for taxi to reach first pixel
        extrataxi = ((0.5 * ((slew_speed**2) / (2 * motor_accel))) / slew_speed) + (
            dwell_time / 2
        )
        taxi_time = accel_time + extrataxi
        # Create np array of times for each pixel in seconds since epoch
        startpixeltime = starttime + taxi_time
        endpixeltime = endtime - taxi_time
        scan_time = endpixeltime - startpixeltime
        timestamps1 = np.linspace(
            startpixeltime, startpixeltime + scan_time, num=len(pixels)
        )
        timestamps = [round(ts, 8) for ts in timestamps1]
        for value, ts in zip(pixels, timestamps):
            yield {
                "data": {self.name: value, self.user_setpoint.name: value},
                "timestamps": {self.name: ts, self.user_setpoint.name: ts},
                "time": ts,
            }

    def describe_collect(self):
        """Describe details for the collect() method"""
        desc = OrderedDict()
        desc.update(self.describe())
        return {"positions": desc}

    def fly(self):
        # Start the trajectory
        destination = self.flyer_taxi_end.get()
        log.debug(f"Flying to {destination}.")
        self.move(destination, wait=True)
        # Wait for the landing
        self.disable_pso()
        self.flying_complete.set(True).wait()
        # Record end time of flight
        self.endtime = time.time()

    def taxi(self):
        # import pdb; pdb.set_trace()
        self.disable_pso()
        # Initalize the PSO
        # Move motor to the scan start point
        self.move(self.pso_start.get(), wait=True)
        # Arm the PSO
        self.enable_pso()
        self.arm_pso()
        # Move the motor to the taxi position
        taxi_start = self.flyer_taxi_start.get()
        log.debug(f"Taxiing to {taxi_start}.")
        self.move(taxi_start, wait=True)
        # Set the speed on the motor
        self.velocity.set(self.flyer_slew_speed.get()).wait()
        # Set timing on the delay for triggering detectors, etc
        self.parent.delay.channel_C.delay.put(0)
        self.parent.delay.output_CD.polarity.put(self.parent.delay.polarities.NEGATIVE)
        # Count-down timer
        # for i in range(10, 0, -1):
        #     print(f"{i}...", end="", flush=True)
        #     time.sleep(1)
        # print("Go!")
        self.ready_to_fly.set(True)

    def stage(self, *args, **kwargs):
        self.ready_to_fly.set(False).wait()
        self.flying_complete.set(False).wait()
        self.starttime = None
        self.endtime = None
        # Save old veolcity to be restored after flying
        self.old_velocity = self.velocity.get()
        super().stage(*args, **kwargs)

    def unstage(self, *args, **kwargs):
        self.velocity.set(self.old_velocity).wait()
        return super().unstage(*args, **kwargs)

    def move(self, position, wait=True, *args, **kwargs):
        motor_status = super().move(position, wait=wait, *args, **kwargs)

        def check_readback(*args, old_value, value, **kwargs) -> bool:
            "Check if taxiing is complete and flying has begun."
            has_arrived = np.isclose(value, position, atol=0.001)
            log.debug(
                f"Checking readback: {value=}, target: {position}, {has_arrived=}"
            )
            return has_arrived

        # Status object is complete motor reaches target value
        readback_status = SubscriptionStatus(self.user_readback, check_readback)
        # Prepare the combined status object
        status = motor_status & readback_status
        if wait:
            status.wait()
        return readback_status

    @property
    def motor_egu_pint(self):
        egu = ureg(self.motor_egu.get())
        return egu

    def flyer_step_size(self):
        """Calculate the size of each step in a fly scan."""
        start_position = self.flyer_start_position.get()
        end_position = self.flyer_end_position.get()
        num_points = self.flyer_num_points.get()
        step_size = abs(start_position - end_position) / (num_points - 1)
        return step_size

    def _update_fly_params(self, *args, **kwargs):
        """Calculate new fly-scan parameters based on signal values.

        Implementation courtesy of Alan Kastengren.

        Computes several parameters describing the fly scan motion.
        These include the actual start position of the motor, the
        actual distance (in encoder counts and distance) between PSO
        pulses, the end position of the motor, and where PSO pulses
        are expected to occcur.  This code ensures that each PSO delta
        is an integer number of encoder counts, since this is how the
        PSO triggering works in hardware.

        These calculations are for MCS scans, where for N bins we need
        N+1 pulses.

        Several fields are set in the class:

        direction
          1 if we are moving positive in user coordinates, −1 if
          negative
        overall_sense
          is our fly motion + or - with respect to encoder counts
        taxi_start
          The starting point for motor movement during flying, accounts
          for needed acceleration of the motor.
        taxi_end
          The target point for motor movement during flying, accounts
          for needed acceleration of the motor.
        pso_start
          The motor position corresponding to the first PSO pulse.
        pso_end
           The motor position corresponding to the last PSO pulse.
        pso_zero
           The motor position where the PSO counter is set to zero.
        encoder_step_size
          The number of encoder counts for each pixel.
        encoder_window_start
          The start of the window within which PSO pulses may be emitted,
          in encoder counts. Should be slightly wider than the actual PSO
          range.
        encoder_window_end
          The end of the window within which PSO pulses may be emitted,
          in encoder counts. Should be slightly wider than the actual PSO
        pixel_positions
          array of places where pixels are, should occur calculated from
          encoder counts then translated to motor positions
        """
        window_buffer = 5
        # Grab any neccessary signals for calculation
        start_position = self.flyer_start_position.get()
        end_position = self.flyer_end_position.get()
        dwell_time = self.flyer_dwell_time.get()
        step_size = self.flyer_step_size()
        encoder_resolution = self.encoder_resolution.get()
        accel_time = self.acceleration.get()
        # Check for sane values
        if dwell_time == 0:
            log.warning(
                f"{self} dwell_time is zero. Could not update fly scan parameters."
            )
            return
        if encoder_resolution == 0:
            log.warning(
                f"{self} encoder resolution is zero. Could not update fly scan"
                " parameters."
            )
            return
        if accel_time <= 0:
            log.warning(
                f"{self} acceleration is non-positive. Could not update fly scan"
                " parameters."
            )
            return
        # Determine the desired direction of travel and overall sense
        # +1 when moving in + encoder direction, -1 if else
        direction = 1 if start_position < end_position else -1
        overall_sense = direction * self.encoder_direction
        # Calculate the step size in encoder steps
        encoder_step_size = round(step_size / encoder_resolution)
        # PSO start/end should be located to where req. start/end are
        # in between steps. Also doubles as the location where slew
        # speed must be met.
        pso_start = start_position - (direction * (step_size / 2))
        pso_end = end_position + (direction * (step_size / 2))
        # Determine taxi distance to accelerate to req speed, v^2/(2*a) = d
        # x1.5 for safety margin
        slew_speed = step_size / dwell_time
        motor_accel = slew_speed / accel_time
        taxi_dist = slew_speed**2 / (2 * motor_accel) * 1.5
        taxi_start = pso_start - (direction * taxi_dist)
        taxi_end = pso_end + (direction * taxi_dist)
        # Actually taxi to the first PSO position before the taxi position
        encoder_taxi_start = (taxi_start - pso_start) / encoder_resolution
        if overall_sense > 0:
            rounder = math.floor
        else:
            rounder = math.ceil
        encoder_taxi_start = (
            rounder(encoder_taxi_start / encoder_step_size) * encoder_step_size
        )
        taxi_start = pso_start + encoder_taxi_start * encoder_resolution
        # Calculate encoder counts within the requested window of the scan
        encoder_window_start = 0  # round(pso_start / encoder_resolution)
        encoder_distance = (pso_end - pso_start) / encoder_resolution
        encoder_window_end = round(encoder_window_start + encoder_distance)
        # Widen the bound a little to make sure we capture the pulse
        encoder_window_start -= overall_sense * window_buffer
        encoder_window_end += overall_sense * window_buffer

        # Check for values outside of the window range for this controller
        def is_valid_window(value):
            window_in_range = self.encoder_window_min < value < self.encoder_window_max
            return window_in_range

        window_range = [encoder_window_start, encoder_window_end]
        encoder_use_window = all([is_valid_window(v) for v in window_range])
        encoder_use_window = encoder_use_window and not self.disable_window.get()
        # Create np array of PSO positions in encoder counts
        _pso_step = encoder_step_size * overall_sense
        _pso_end = encoder_distance + 0.5 * _pso_step
        encoder_pso_positions = np.arange(0, _pso_end, _pso_step)
        # Transform from PSO positions from encoder counts to engineering units
        pso_positions = (encoder_pso_positions * encoder_resolution) + pso_start
        # Tranforms from pulse positions to pixel centers
        pixel_positions = (pso_positions[1:] + pso_positions[:-1]) / 2
        # Set all the calculated variables
        [
            stat.wait()
            for stat in [
                self.encoder_step_size.set(encoder_step_size),
                self.pso_start.set(pso_start),
                self.pso_end.set(pso_end),
                self.flyer_slew_speed.set(slew_speed),
                self.flyer_taxi_start.set(taxi_start),
                self.flyer_taxi_end.set(taxi_end),
                self.encoder_window_start.set(encoder_window_start),
                self.encoder_window_end.set(encoder_window_end),
                self.encoder_use_window.set(encoder_use_window),
            ]
        ]
        self.encoder_pso_positions = encoder_pso_positions
        self.pso_positions = pso_positions
        self.pixel_positions = pixel_positions

    def send_command(self, cmd: str):
        """Send commands directly to the aerotech ensemble controller.

        Returns
        =======
        status
          The Ophyd status object for this write.

        """
        status = self.parent.asyn.ascii_output.set(cmd, settle_time=0.1)
        status.wait()
        return status

    def disable_pso(self):
        for axis in range(2):
            self.send_command(f"PSOCONTROL @{axis} OFF")

    def check_flyscan_bounds(self):
        """Check that the fly-scan params are sane at the scan start and end.

        This checks to make sure no spurious pulses are expected from taxiing.

        """
        end_points = [
            (self.flyer_taxi_start, self.pso_start),
            (self.flyer_taxi_end, self.pso_end),
        ]
        step_size = self.flyer_step_size()
        for taxi, pso in end_points:
            # Make sure we're not going to have extra pulses
            taxi_distance = abs(taxi.get() - pso.get())
            if taxi_distance > (1.1 * step_size):
                raise InvalidScanParameters(
                    f"Scan parameters for {taxi}, {pso}, {self.flyer_step_size} would produce"
                    " extra pulses without a window."
                )

    def enable_pso(self):
        num_axis = 1
        use_window = self.encoder_use_window.get()
        if not use_window:
            self.check_flyscan_bounds()
        # Erase any previous PSO control
        self.send_command(f"PSOCONTROL {self.axis} RESET")
        # Set the output to occur from the I/O terminal on the
        # controller
        self.send_command(f"PSOOUTPUT {self.axis} CONTROL {num_axis}")
        # Set a pulse 10 us long, 20 us total duration, so 10 us
        # on, 10 us off
        self.send_command(f"PSOPULSE {self.axis} TIME 20, 10")
        # Set the pulses to only occur in a specific window
        if use_window:
            self.send_command(f"PSOOUTPUT {self.axis} PULSE WINDOW MASK")
        else:
            self.send_command(f"PSOOUTPUT {self.axis} PULSE")
        # Set which encoder we will use.  3 = the MXH (encoder
        # multiplier) input. For Ensemble lab, 6 is horizontal encoder
        self.send_command(f"PSOTRACK {self.axis} INPUT {self.encoder}")
        # Set the distance between pulses in encoder counts
        self.send_command(
            f"PSODISTANCE {self.axis} FIXED {self.encoder_step_size.get()}"
        )
        # Which encoder is being used to calculate whether we are
        # in the window.
        if use_window:
            self.send_command(f"PSOWINDOW {self.axis} {num_axis} INPUT {self.encoder}")
            # Calculate window function parameters. Must be in encoder
            # counts, and is referenced from the stage location where
            # we arm the PSO
            window_range = (
                self.encoder_window_start.get(),
                self.encoder_window_end.get(),
            )
            self.send_command(
                f"PSOWINDOW {self.axis} {num_axis} RANGE "
                f"{min(window_range)},{max(window_range)}"
            )

    def arm_pso(self):
        self.send_command(f"PSOCONTROL {self.axis} ARM")


class AerotechStage(XYStage):
    """An XY stage for an Aerotech stage with fly-scanning capabilities.

    Parameters
    ==========

    pv_vert
      The suffix to the PV for the vertical motor.
    pv_horiz
      The suffix to the PV for the horizontal motor.
    """

    horiz = FCpt(
        AerotechFlyer,
        "{prefix}{pv_horiz}",
        axis="@0",
        encoder=6,
        labels={"motors", "flyers"},
    )
    vert = FCpt(
        AerotechFlyer,
        "{prefix}{pv_vert}",
        axis="@1",
        encoder=7,
        labels={"motors", "flyers"},
    )
    asyn = Cpt(AsynRecord, ":asynEns", name="async", labels={"asyns"})
    # A digital delay generator for providing a gate signal
    delay = FCpt(DG645Delay, "{delay_prefix}:", kind=Kind.config)

    def __init__(self, *args, delay_prefix, **kwargs):
        self.delay_prefix = delay_prefix
        super().__init__(*args, **kwargs)


def load_aerotech_stages(config=None):
    """Load Aerotech XY stages defined in the configuration files'
    ``[aerotech_stage]`` sections.

    """
    if config is None:
        config = load_config()
    devices = []
    for name, stage_data in config.get("aerotech_stage", {}).items():
        devices.append(
            make_device(
                AerotechStage,
                name=name,
                prefix=stage_data["prefix"],
                delay_prefix=stage_data["delay_prefix"],
                pv_vert=stage_data["pv_vert"],
                pv_horiz=stage_data["pv_horiz"],
            )
        )
    return devices


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from firefly.slits import SlitsDisplay

# from haven.instrument import mirrors


class KBMirrorsDisplay(SlitsDisplay):
    @property
    def caqtdm_ui_file(self):
        # Go up the class list until we find a class that is recognized
        if self.device.horiz.bendable or self.device.vert.bendable:
            ui_file = "/net/s25data/xorApps/ui/KB_mirrors_and_benders.ui"
        else:
            ui_file = "/net/s25data/xorApps/ui/KB_mirrors.ui"
        return ui_file

    def ui_filename(self):
        return "kb_mirrors.ui"

    def customize_ui(self):
        # Enable/disable bender controls
        horiz = self.device.horiz
        self.ui.horizontal_upstream_display.setEnabled(horiz.bendable)
        self.ui.horizontal_downstream_display.setEnabled(horiz.bendable)
        vert = self.device.vert
        self.ui.vertical_upstream_display.setEnabled(vert.bendable)
        self.ui.vertical_downstream_display.setEnabled(vert.bendable)

    def launch_caqtdm(self):
        # Sort out the prefix from the slit designator
        prefix = self.device.prefix.strip(":")
        pieces = prefix.split(":")
        # Build the macros for the caQtDM panels
        P = ":".join(pieces[:-1])
        P = f"{P}:"
        KB = pieces[-1]
        KBH = self.device.horiz.prefix.replace(P, "").strip(":")
        KBV = self.device.vert.prefix.replace(P, "").strip(":")

        def suffix(signal):
            return signal.prefix.split(":")[-1]

        caqtdm_macros = {
            "P": f"{P}",
            "PM": P,
            "KB": KB,
            "KBH": KBH,
            "KBV": KBV,
            # Macros for the real motors
            "KBHUS": suffix(self.device.horiz.upstream),
            "KBHDS": suffix(self.device.horiz.downstream),
            "KBVUS": suffix(self.device.vert.upstream),
            "KBVDS": suffix(self.device.vert.downstream),
            # Macros for the transform records
            "KB1": KBH.replace(":", ""),
            "KB2": KBV.replace(":", ""),
        }
        # Macros for each mirror's bender motors
        horiz = self.device.horiz
        if horiz.bendable:
            caqtdm_macros.update(
                {
                    "HBUS": suffix(horiz.bender_upstream),
                    "HBDS": suffix(horiz.bender_downstream),
                }
            )
        vert = self.device.vert
        if vert.bendable:
            caqtdm_macros.update(
                {
                    "VBUS": suffix(vert.bender_upstream),
                    "VBDS": suffix(vert.bender_downstream),
                }
            )
        # Launch the caQtDM panel
        super(SlitsDisplay, self).launch_caqtdm(macros=caqtdm_macros)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from haven.instrument import aps


def test_load_aps(sim_registry):
    aps.load_aps()
    aps_ = sim_registry.find(name="APS")
    assert hasattr(aps_, "current")


def test_read_attrs():
    device = aps.ApsMachine(name="Aps")
    read_attrs = ["current", "lifetime"]
    for attr in read_attrs:
        assert attr in device.read_attrs


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from haven.instrument import Table, load_tables


def test_vertical_table():
    """A table that can only move up and down."""
    tbl = Table("255idcVME:", vertical_motor="m3", name="my_table")
    # Check that the correct components were created
    assert hasattr(tbl, "vertical")
    assert tbl.vertical.prefix == "255idcVME:m3"
    assert not hasattr(tbl, "horizontal")
    assert not hasattr(tbl, "upstream")
    assert not hasattr(tbl, "downstream")
    assert not hasattr(tbl, "pitch")
    assert not hasattr(tbl, "vertical_drive_transform")
    assert not hasattr(tbl, "vertical_readback_transform")


def test_horizontal_table():
    """A table that can only move left and right."""
    tbl = Table("255idcVME:", horizontal_motor="m3", name="my_table")
    # Check that the correct components were created
    assert not hasattr(tbl, "vertical")
    assert hasattr(tbl, "horizontal")
    assert tbl.horizontal.prefix == "255idcVME:m3"
    assert not hasattr(tbl, "upstream")
    assert not hasattr(tbl, "downstream")
    assert not hasattr(tbl, "pitch")
    assert not hasattr(tbl, "vertical_drive_transform")
    assert not hasattr(tbl, "vertical_readback_transform")


def test_pitch_table():
    """A table that can move vertical and adjust the angle."""
    tbl = Table(
        "255idcVME:",
        pseudo_motors="table_us:",
        transforms="table_us_trans:",
        upstream_motor="m3",
        downstream_motor="m4",
        name="my_table",
    )
    # Check that the correct components were created (or not)
    assert not hasattr(tbl, "horizontal")
    assert hasattr(tbl, "upstream")
    assert tbl.upstream.prefix == "255idcVME:m3"
    assert hasattr(tbl, "downstream")
    assert tbl.downstream.prefix == "255idcVME:m4"
    assert hasattr(tbl, "vertical")
    assert tbl.vertical.prefix == "255idcVME:table_us:height"
    assert hasattr(tbl, "pitch")
    assert tbl.pitch.prefix == "255idcVME:table_us:pitch"
    # Check the transforms
    assert hasattr(tbl, "vertical_drive_transform")
    assert tbl.vertical_drive_transform.prefix == "255idcVME:table_us_trans:Drive"
    assert hasattr(tbl, "vertical_readback_transform")
    assert tbl.vertical_readback_transform.prefix == "255idcVME:table_us_trans:Readback"


def test_load_tables(sim_registry):
    load_tables()
    # Check that the vertical/horizontal table has the right motors
    table = sim_registry.find(name="upstream_table")
    assert isinstance(table, Table)
    assert not hasattr(table, "upstream")
    assert not hasattr(table, "downstream")
    assert table.vertical.prefix == "255idcVME:m26"
    assert table.horizontal.prefix == "255idcVME:m25"
    # Check that the 2-leg table has the right motors
    table = sim_registry.find(name="downstream_table")
    assert isinstance(table, Table)
    assert table.upstream.prefix == "255idcVME:m21"
    assert table.downstream.prefix == "255idcVME:m22"
    assert table.vertical.prefix == "255idcVME:table_ds:height"
    assert not hasattr(table, "horizontal")



from apstools.synApps import EpicsRecordDeviceCommonAll, EpicsRecordInputFields
from ophyd import Device, EpicsSignal
from ophyd import FormattedComponent as FCpt
from strenum import StrEnum


class AnalogInput(EpicsRecordInputFields, EpicsRecordDeviceCommonAll):
    class DiffStates(StrEnum):
        single_ended = "Single-Ended"
        differential = "Differential"

    differential = FCpt(
        EpicsSignal, "{self.base_prefix}Diff{self.ch_num}", kind="config"
    )
    high = FCpt(EpicsSignal, "{self.base_prefix}HOPR{self.ch_num}", kind="config")
    low = FCpt(EpicsSignal, "{self.base_prefix}LOPR{self.ch_num}", kind="config")
    temperature_units = FCpt(
        EpicsSignal,
        "{self.base_prefix}TempUnits{self.ch_num}",
        kind="config",
    )
    resolution = FCpt(
        EpicsSignal, "{self.base_prefix}Resolution{self.ch_num}", kind="config"
    )
    range = FCpt(
        EpicsSignal,
        "{self.base_prefix}Range{self.ch_num}",
        kind="config",
    )
    mode = FCpt(EpicsSignal, "{self.base_prefix}Mode{self.ch_num}", kind="config")
    enable = FCpt(EpicsSignal, "{self.base_prefix}Enable{self.ch_num}")

    @property
    def base_prefix(self):
        return self.prefix.rstrip("0123456789")

    @property
    def ch_num(self):
        return self.prefix[len(self.base_prefix) :]


class LabJackT7(Device):
    """A labjack T7 data acquisition unit (DAQ)."""

    ...


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import warnings
from pathlib import Path
from typing import Mapping

import haven
from firefly import display
from haven.instrument import Table


class CaQtDMBase:
    """caQtDM parameters for the given table.

    Different table geometries require different caQtDM
    configurations.

    """

    ui_file: str = ""
    macros: Mapping = {}
    table: Table

    def __init__(self, table: Table):
        self.table = table

    def motor_attrs(self):
        attr_names = ["horizontal", "vertical"]
        # Look for which motor is present on this table device
        attrs = []
        for attr in attr_names:
            if getattr(self.table, attr, None) is not None:
                attrs.append(attr)
        return attrs


class TwoLegCaQtDM(CaQtDMBase):
    """caQtDM parameters for a table with two ."""

    ui_file = "/net/s25data/xorApps/ui/table_2leg.ui"

    @property
    def macros(self):
        macros = {
            "PM": self.table.prefix,
            "TB": self.table.pseudo_motors.strip(":"),
            "TR": self.table.transforms.strip(":"),
            "TBUS": self.table.__class__.upstream.suffix,
            "TBDS": self.table.__class__.downstream.suffix,
        }
        # See if there's a horizontal motor
        horizontal = getattr(self.table.__class__, "horizontal", None)
        if horizontal is not None:
            macros["TBH"] = horizontal.suffix
        return macros


class SingleMotorCaQtDM(CaQtDMBase):
    """caQtDM parameters for a table with only a single motor."""

    ui_file = "/APSshare/epics/synApps_6_2_1/support/motor-R7-2-2//motorApp/op/ui/autoconvert/motorx.ui"

    @property
    def macros(self):
        # Look for which motor is present on this table device
        for attr in self.motor_attrs():
            component = getattr(self.table.__class__, attr)
            return {"M": component.suffix}
        raise RuntimeError(f"Could not find single motor for {self.device}.")


class MultipleMotorCaQtDM(CaQtDMBase):
    """caQtDM parameters for a table with only multiple independent motors."""

    @property
    def ui_file(self):
        ui_dir = Path(
            "/APSshare/epics/synApps_6_2_1/support/motor-R7-2-2//motorApp/op/ui/autoconvert/"
        )
        num_motors = len(self.motor_attrs())
        ui_path = ui_dir / f"motor{num_motors}x.ui"
        return str(ui_path)

    @property
    def macros(self):
        # Look for which motor is present on this table device
        macros = {}
        for idx, name in enumerate(self.motor_attrs()):
            component = getattr(self.table.__class__, name)
            key = f"M{idx+1}"
            macros[key] = component.suffix
        return macros


class TableDisplay(display.FireflyDisplay):
    def customize_device(self):
        self.device = haven.registry.find(self.macros()["DEVICE"])
        # Determine which flavor of caQtDM parameters we need
        if self.num_legs == 2:
            self.caqtdm = TwoLegCaQtDM(table=self.device)
        elif self.num_motors == 1:
            self.caqtdm = SingleMotorCaQtDM(table=self.device)
        elif self.num_motors > 1:
            self.caqtdm = MultipleMotorCaQtDM(table=self.device)
        else:
            warnings.warn(
                f"Could not determine caQtDM parameters for device: {self.device}."
            )
            self.caqtdm = CaQtDMBase(table=self.device)

    def ui_filename(self):
        return "table.ui"

    @property
    def caqtdm_ui_file(self):
        return self.caqtdm.ui_file

    @property
    def num_legs(self):
        """How motorized legs does this table have?

        If 1, it's a simple motorized table. If greater than 1, it
        will probably have some angular control over the table
        surface.

        """
        leg_names = {"upstream", "downstream"}
        num_legs = len(list(set(self.device.component_names) & leg_names))
        return num_legs

    @property
    def num_motors(self):
        """How many motors does this table have?

        Does not include the pseudo motors produce through the
        sum2Diff EPICS record.

        """
        motor_names = {"horizontal", "vertical", "upstream", "downstream"}
        num_motors = len(list(set(self.device.component_names) & motor_names))
        return num_motors

    def customize_ui(self):
        # Disable motor controls if the given axis is not available
        self.ui.pitch_embedded_display.setEnabled(hasattr(self.device, "pitch"))
        self.ui.vertical_embedded_display.setEnabled(hasattr(self.device, "vertical"))
        self.ui.horizontal_embedded_display.setEnabled(
            hasattr(self.device, "horizontal")
        )

    def launch_caqtdm(self):
        # Sort out the prefix from the slit designator
        prefix = self.device.prefix.strip(":")
        pieces = prefix.split(":")
        # Build the macros for the caQtDM panels
        P = ":".join(pieces[:-1])
        caqtdm_macros = dict(
            P=self.device.prefix,
            **self.caqtdm.macros,
        )
        # Launch the caQtDM panel
        super().launch_caqtdm(macros=caqtdm_macros)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from haven import registry
from haven.instrument import shutter


def test_shutter(sim_registry, beamline_connected):
    shutter.load_shutters()
    shutters = list(registry.findall(label="shutters"))
    assert len(shutters) == 2
    shutterA = registry.find(name="front_end_shutter")
    assert shutterA.name == "front_end_shutter"
    assert shutterA.open_signal.pvname == "PSS:99ID:FES_OPEN_EPICS.VAL"
    assert shutterA.close_signal.pvname == "PSS:99ID:FES_CLOSE_EPICS.VAL"
    assert shutterA.pss_state.pvname == "PSS:99ID:A_BEAM_PRESENT"
    shutterC = registry.find(name="hutch_shutter")
    assert shutterC.name == "hutch_shutter"
    assert shutterC.open_signal.pvname == "PSS:99ID:SCS_OPEN_EPICS.VAL"
    assert shutterC.close_signal.pvname == "PSS:99ID:SCS_CLOSE_EPICS.VAL"
    assert shutterC.pss_state.pvname == "PSS:99ID:C_BEAM_PRESENT"


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import pytest
from ophydregistry.exceptions import ComponentNotFound

from haven.instrument import Robot, load_robots


def test_robot():
    robot = Robot("25idAustin", name="robot")
    # Check PVs are correct
    assert robot.i.user_readback.pvname == "25idAustin:i.RBV"
    assert robot.samples.sample8.present.pvname == "25idAustin:sample8:present"
    assert robot.samples.sample8.rz.pvname == "25idAustin:sample8:rz"


def test_load_robot(sim_registry):
    load_robots()
    # Test the robot info is extracted properly
    rbt = sim_registry.find(label="robots")
    assert rbt.name == "A"
    assert rbt.prefix == "255idAustin"


def test_load_no_robot(sim_registry):
    load_robots(config={})

    # Test the robot is not in config
    result = pytest.raises(ComponentNotFound, sim_registry.findall, label="robots")

    # Assert that the expected exception is raised
    assert 'Could not find components matching: label="robots", name="None"' in str(
        result.value
    )


# -----------------------------------------------------------------------------
# :author:    Yanna Chen
# :email:     yannachen@anl.gov
# :copyright: Copyright © 2024, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import pytest
from ophyd import PVPositioner
from ophyd.sim import instantiate_fake_device

from haven.instrument.energy_positioner import EnergyPositioner, load_energy_positioner


@pytest.fixture()
def positioner():
    positioner = instantiate_fake_device(
        EnergyPositioner,
        name="energy",
        mono_prefix="255idMono:",
        undulator_prefix="S255ID:",
    )
    positioner.monochromator.energy._use_limits = False
    positioner.monochromator.energy.user_setpoint._use_limits = False
    return positioner


def test_set_energy(positioner):
    # Set up dependent values
    positioner.monochromator.id_offset.set(150).wait(timeout=3)
    # Change the energy
    positioner.set(10000, timeout=3)
    # Check that all the sub-components were set properly
    assert positioner.monochromator.energy.get().user_setpoint == 10000
    assert positioner.undulator.energy.get().setpoint == 10.150


def test_load_energy_positioner(sim_registry):
    load_energy_positioner()
    energy = sim_registry["energy"]
    assert isinstance(energy, PVPositioner)
    assert hasattr(energy, "monochromator")
    assert hasattr(energy, "undulator")


def test_real_to_pseudo_positioner(positioner):
    positioner.monochromator.energy.user_readback._readback = 5000.0
    # Check that the pseudo single is updated
    assert positioner.get(use_monitor=False).readback == 5000.0


def test_energy_limits(positioner):
    """Check that the energy range is determined by the limits of the
    dependent signals.

    """
    positioner.monochromator.energy.user_setpoint.sim_set_limits((0, 35000))
    positioner.undulator.energy.setpoint.sim_set_limits((0.01, 1000))  # (10, 1000000)
    assert positioner.limits == (10, 35000)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from haven.instrument.mirrors import (
    BendableHighHeatLoadMirror,
    BendableKBMirror,
    HighHeatLoadMirror,
    KBMirror,
    KBMirrors,
    load_mirrors,
)


def test_high_heat_load_mirror_PVs():
    mirror = BendableHighHeatLoadMirror(prefix="255ida:ORM2:", name="orm2")
    # Check the motor PVs
    assert mirror.transverse.user_readback.pvname == "255ida:ORM2:m1.RBV"
    assert mirror.roll.user_setpoint.pvname == "255ida:ORM2:m2.VAL"
    assert mirror.upstream.user_setpoint.pvname == "255ida:ORM2:m3.VAL"
    assert mirror.downstream.user_setpoint.pvname == "255ida:ORM2:m4.VAL"
    assert mirror.normal.user_setpoint.pvname == "255ida:ORM2:lateral.VAL"
    assert mirror.pitch.user_setpoint.pvname == "255ida:ORM2:coarsePitch.VAL"
    assert mirror.bender.user_setpoint.pvname == "255ida:ORM2:m5.VAL"
    # Check the transform PVs
    assert (
        mirror.drive_transform.channels.B.input_pv.pvname
        == "255ida:ORM2:lats:Drive.INPB"
    )
    assert (
        mirror.readback_transform.channels.B.input_pv.pvname
        == "255ida:ORM2:lats:Readback.INPB"
    )


def test_kb_mirrors_PVs():
    kb = KBMirrors(
        prefix="255idcVME:LongKB_Cdn:",
        horiz_upstream_motor="255idcVME:m33",
        horiz_downstream_motor="255idcVME:m34",
        vert_upstream_motor="255idcVME:m35",
        vert_downstream_motor="255idcVME:m36",
        name="kb",
    )
    # assert not kb.connected
    assert isinstance(kb.vert, KBMirror)
    assert isinstance(kb.horiz, KBMirror)
    # Check PVs
    assert kb.vert.prefix == "255idcVME:LongKB_Cdn:V:"
    # "25idcVME:LongKB_Cdn:H:pitch.VAL"
    assert kb.horiz.pitch.user_setpoint.pvname == "255idcVME:LongKB_Cdn:H:pitch.VAL"
    assert kb.vert.normal.user_setpoint.pvname == "255idcVME:LongKB_Cdn:V:height.VAL"
    assert kb.horiz.upstream.user_setpoint.pvname == "255idcVME:m33.VAL"
    assert kb.horiz.downstream.user_setpoint.pvname == "255idcVME:m34.VAL"
    assert kb.vert.upstream.user_setpoint.pvname == "255idcVME:m35.VAL"
    assert kb.vert.downstream.user_setpoint.pvname == "255idcVME:m36.VAL"
    # Check the transforms
    assert (
        kb.horiz.drive_transform.channels.B.input_pv.pvname
        == "255idcVME:LongKB_CdnH:Drive.INPB"
    )
    assert (
        kb.horiz.readback_transform.channels.B.input_pv.pvname
        == "255idcVME:LongKB_CdnH:Readback.INPB"
    )


def test_load_mirrors(sim_registry):
    load_mirrors()
    # Check that the KB mirrors were created
    kb_mirrors = sim_registry.find(name="KB")
    assert isinstance(kb_mirrors, KBMirrors)
    assert not isinstance(kb_mirrors.horiz, BendableKBMirror)
    assert not isinstance(kb_mirrors.vert, BendableKBMirror)
    # Check that the KB mirrors selects the bendable version
    kb_mirrors = sim_registry.find(name="LongKB_Cdn")
    assert isinstance(kb_mirrors, KBMirrors)
    assert kb_mirrors.horiz.bendable
    assert kb_mirrors.vert.bendable
    assert hasattr(kb_mirrors.horiz, "bender_upstream")
    assert hasattr(kb_mirrors.horiz, "bender_downstream")
    # Check that the HHL mirrors were created
    hhl_mirrors = sim_registry.find(name="ORM1")
    assert isinstance(hhl_mirrors, HighHeatLoadMirror)
    assert not isinstance(hhl_mirrors, BendableHighHeatLoadMirror)
    # Check that the HHL mirror selects the bendable version
    hhl_mirrors = sim_registry.find(name="ORM2")
    assert isinstance(hhl_mirrors, BendableHighHeatLoadMirror)



import logging

import qtawesome as qta
from pydm.widgets import PyDMByteIndicator, PyDMPushButton
from qtpy.QtCore import Signal
from qtpy.QtWidgets import QHBoxLayout, QSizePolicy

from firefly import display
from haven import registry
from haven.instrument.xia_pfcu import ShutterStates

log = logging.getLogger(__name__)


def name_to_title(name: str):
    """Convert a python-valid Ophyd object name to a human-readable
    title-case string.

    """
    title = name.replace("_", " ").title()
    return title


class StatusDisplay(display.FireflyDisplay):
    caqtdm_ui_file: str = "/net/s25data/xorApps/ui/25id_main.ui"
    bss_window_requested = Signal()

    def add_shutter_widgets(self):
        form = self.ui.beamline_layout
        # Remove existing layouts
        form.removeRow(self.ui.shutter_A_layout)
        form.removeRow(self.ui.shutter_CD_layout)
        # Add widgets for shutters
        shutters = registry.findall("shutters", allow_none=True)
        row_idx = 4
        on_color = self.ui.shutter_permit_indicator.onColor
        off_color = self.ui.shutter_permit_indicator.offColor
        for shutter in shutters[::-1]:
            # Add a layout with the buttons
            layout = QHBoxLayout()
            name = shutter.attr_name if shutter.attr_name != "" else shutter.name
            label = name_to_title(name) + ":"
            form.insertRow(row_idx, label, layout)
            # Indicator to show if the shutter is open
            indicator = PyDMByteIndicator(
                parent=self, init_channel=f"haven://{shutter.name}.readback"
            )
            # indicator.showLabels = False
            indicator.labels = ["Closed", "Half Open"]
            indicator.numBits = 2
            indicator.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
            # Switch colors because open is 0 which should means "good"
            indicator.offColor = on_color
            indicator.onColor = off_color
            layout.addWidget(indicator)
            # Button to open the shutter
            open_btn = PyDMPushButton(
                parent=self,
                label="Open",
                icon=qta.icon("mdi.window-shutter-open"),
                pressValue=ShutterStates.OPEN,
                relative=False,
                init_channel=f"haven://{shutter.name}.setpoint",
            )
            layout.addWidget(open_btn)
            # Button to close the shutter
            close_btn = PyDMPushButton(
                parent=self,
                label="Close",
                icon=qta.icon("mdi.window-shutter"),
                pressValue=ShutterStates.CLOSED,
                relative=False,
                init_channel=f"haven://{shutter.name}.setpoint",
            )
            layout.addWidget(close_btn)

    def customize_ui(self):
        self.ui.bss_modify_button.clicked.connect(self.bss_window_requested.emit)
        self.ui.bss_modify_button.setIcon(qta.icon("fa5s.calendar"))
        self.add_shutter_widgets()

    def ui_filename(self):
        return "status.ui"


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import datetime as dt
import logging
import re
import unicodedata
import warnings
from pathlib import Path
from typing import Mapping, Optional, Sequence, Union

from bluesky.callbacks import CallbackBase

from . import exceptions, registry

log = logging.getLogger(__name__)


def slugify(value, allow_unicode=False):
    """
    Taken from https://github.com/django/django/blob/master/django/utils/text.py
    Convert to ASCII if 'allow_unicode' is False. Convert spaces or repeated
    dashes to single dashes. Remove characters that aren't alphanumerics,
    underscores, or hyphens. Also strip leading and
    trailing whitespace, dashes, and underscores.
    """
    value = str(value)
    if allow_unicode:
        value = unicodedata.normalize("NFKC", value)
    else:
        value = (
            unicodedata.normalize("NFKD", value)
            .encode("ascii", "ignore")
            .decode("ascii")
        )
    return re.sub(r"[-\s]+", "-", value).strip("-_")


class XDIWriter(CallbackBase):
    """A callback class for bluesky that will save data to an XDI file.

    The XAFS data interchange (XDI) format is a plain text,
    tab-separated file format that standardizes XAFS data. Metadata is
    also included in headers.

    The location of the file is determined by passing the *filename*
    parameter in when created the writer object:

    .. code-block:: python

      writer = XDIWriter(fd="./my_experiment.xdi")

    Placeholders can be included that will be filled-in with metadata
    during the start phase of the plan, for example:

    .. code-block:: python

      plan = energy_scan(..., E0="Ni_K", md=dict(sample_name="nickel oxide"))
      writer_callback = XDIWriter(fd="./{year}{month}{day}_{sample_name}_{edge}.xdi")
      RE(plan, writer)

    Assuming the date is 2022-08-19, then the filename will become
    "20220819_nickel-oxide_Ni_K.xdi".

    *{year}*, *{month}*, *{day}* describe the numerical, zero-padded
    year, month and day when the callback handles the start document
    for the plan. *{short_uid}* is the portion of the scan UID up to
    the first '-' character. *{manager_path}* will be the full path
    specified by the device labeled "beamline_manager" that has a
    ``local_storage.full_path`` signal. The remaining fields in the
    start document are also available, and will likely vary from plan
    to plan. Either consult the documentation for the plan being
    executed, or inspect the logs for the dictionary metadata at level
    ``logging.INFO`` when using an invalid placeholder.

    Parameters
    ==========
    fd
      Either an open file with write intent, or a Path or string with
      the file location. If *fd* is not already an open file, it will
      be created when the ``start()`` method is called.
    auto_directory
      If true, the current local storage directory will be prepended
      to *fd* if *fd* is not an open file object. The directory is
      determined from the
      :py:class:`~haven.instrument.beamline_manager.BeamlineManager`
      object.

    """

    fp: Optional[Union[str, Path]] = None
    column_names: Optional[Sequence[str]] = None
    start_time: dt.datetime = None
    stream_name: str = "primary"

    start_doc: Mapping = None
    descriptor_doc: Mapping = None

    _fd = None
    _fp_template: Optional[Union[str, Path]] = None
    _last_uid: str = ""
    _primary_uid: str
    _primary_descriptor: Mapping
    _secondary_uids: Sequence[str]

    def __init__(self, fd: Union[Path, str], *args, **kwargs):
        is_file_obj = hasattr(fd, "writable")
        if is_file_obj:
            # *fd* is an open file object
            # Make sure it is writable
            if fd.writable():
                log.debug(f"Found open, writable file: {fd}")
                self._fd = fd
            else:
                msg = f"No write intent on file: {fd}"
                log.error(msg)
                raise exceptions.FileNotWritable(msg)
        else:
            # Assume *fd* is a path to a file
            self._fp_template = Path(fd).expanduser()
            self.fp = self._fp_template
        return super().__init__(*args, **kwargs)

    @property
    def fd(self):
        """Retrieve the open writable file object for this writer."""
        if self._fd is None:
            log.debug(f"Opening file: {self.fp.resolve()}")
            self.fp.parent.mkdir(exist_ok=True, parents=True)
            self._fd = open(self.fp, mode="x")
        return self._fd

    def _path_metadata(self, doc={}):
        """Prepare the metadata for string formatting the output file path."""
        md = {
            "year": self.start_time.strftime("%Y"),
            "month": self.start_time.strftime("%m"),
            "day": self.start_time.strftime("%d"),
        }
        # Add a shortened version of the UID
        try:
            md["short_uid"] = doc["uid"].split("-")[0]
        except KeyError:
            pass
        # Add local storage directory
        try:
            manager_name = "beamline_manager.local_storage.full_path"
            md["manager_path"] = registry[manager_name].get(as_string=True)
        except exceptions.ComponentNotFound:
            log.warning(f"Could not find beamline manager {manager_name}")
        md.update(doc)
        return md

    def start(self, doc):
        self.start_time = dt.datetime.now().astimezone()
        self.column_names = None
        self._primary_uid = None
        self._secondary_uids = []
        # Format the file name based on metadata
        is_new_uid = doc.get("uid", "") != self._last_uid
        if self._fp_template is not None:
            # Make sure any previous runs are closed
            if is_new_uid:
                self.close()
            fp = str(self._fp_template)
            md = self._path_metadata(doc=doc)
            try:
                fp = fp.format(**md)
            except KeyError as e:
                msg = f"Could not find match key {e} in {fp}."
                log.error(msg)
                log.info(f"Metadata is: {md}")
                raise exceptions.XDIFilenameKeyNotFound(msg) from None
            fp = slugify(fp)
            self.fp = Path(fp)
        # Save the rest of the start doc so we can write the headers
        # when we get our first datum
        self.start_doc = doc
        self._last_uid = doc.get("uid", "")
        # Open the file, just to be sure we can
        self.fd

    def stop(self, doc):
        self.close()

    def descriptor(self, doc):
        if doc["name"] == self.stream_name:
            self._primary_uid = doc["uid"]
            self._primary_descriptor = doc
        else:
            self._secondary_uids.append(doc["uid"])
            return
        # Determine columns to use for storing events later
        if self.column_names is None:
            # Get column names from the scan hinted signals
            names = [val["fields"] for val in doc["hints"].values()]
            names = [n for sublist in names for n in sublist]
            # Sort column names so that energy-related fields are first
            names = sorted(names, key=lambda x: not x.startswith("energy"))
            self.column_names = names + ["time"]
            self.write_header(self.start_doc)
        # Save the descriptor doc for later
        self.descriptor_doc = doc

    def close(self):
        """Ensure any open files are closed."""
        try:
            self._fd.close()
        except ValueError:
            log.debug(f"Could not close file: {self.fd}")
        except AttributeError:
            log.debug(f"No file to close, skipping.")
        else:
            self._fd = None

    def write_header(self, doc):
        fd = self.fd
        # Write package version information
        versions = ["XDI/1.0"]
        versions += [f"{name}/{ver}" for name, ver in doc.get("versions", {}).items()]
        fd.write(f"# {' '.join(versions)}\n")
        # Column Names
        columns = [
            f"# Column.{num+1}: {name}\n" for num, name in enumerate(self.column_names)
        ]
        fd.write("".join(columns))
        # X-ray edge information
        edge_str = doc.get("edge", None)
        try:
            elem, edge = edge_str.split("_")
        except (AttributeError, ValueError):
            msg = f"Could not parse X-ray edge metadata: {edge_str}"
            warnings.warn(msg)
            log.warning(msg)
        else:
            fd.write(f"# Element.symbol: {elem}\n")
            fd.write(f"# Element.edge: {edge}\n")
        # Monochromator information
        try:
            d_spacing = doc["d_spacing"]
        except KeyError:
            pass
        else:
            fd.write(f"# Mono.d_spacing: {d_spacing}\n")
        # Facility information
        now = self.start_time
        fd.write(f"# Scan.start_time: {now.strftime('%Y-%m-%d %H:%M:%S%z')}\n")
        md_paths = [
            "facility.name",
            "facility.xray_source",
            "beamline.name",
            "beamline.pv_prefix",
        ]
        for path in md_paths:
            val = doc
            try:
                for piece in path.split("."):
                    val = val[piece]
                fd.write(f"# {path}: {val}\n")
            except KeyError:
                continue
        # Scan ID
        fd.write(f"# uid: {self.start_doc.get('uid', '')}\n")
        # Header end token
        fd.write("# -------------\n")

    def event(self, doc):
        """Save the data in tab-separated value, in order of
        ``self.column_names``.

        """
        # Ignore non-primary data streams
        if doc["descriptor"] in self._secondary_uids:
            return
        elif doc["descriptor"] != self._primary_uid:
            # We're getting data out of order, so we can't save to the file
            msg = (
                "No descriptor document available. "
                "The descriptor document should be passed to "
                f"{self}.descriptor() before providing event documents."
            )
            raise exceptions.DocumentNotFound(msg)
        # Read in and store the actual data
        data = doc["data"]
        fd = self.fd
        values = []
        for col in self.column_names:
            if col == "time":
                values.append(str(doc["time"]))
            else:
                try:
                    values.append(str(data[col]))
                except KeyError:
                    msg = f"Could not find signal {col} in datum."
                    log.warning(msg)
                    warnings.warn(msg)
        line = "\t".join(values)
        fd.write(line + "\n")


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import logging
import warnings

import haven
from firefly import display
from haven.instrument import slits

log = logging.getLogger(__name__)


class SlitsDisplay(display.FireflyDisplay):
    caqtdm_ui_filenames = {
        slits.BladeSlits: "/APSshare/epics/synApps_6_2_1/support/optics-R2-13-5//opticsApp/op/ui/autoconvert/4slitGraphic.ui",
        slits.ApertureSlits: "/net/s25data/xorApps/epics/synApps_6_2/ioc/25ida/25idaApp/op/ui/maskApertureSlit.ui",
    }

    def customize_device(self):
        self.device = haven.registry.find(self.macros()["DEVICE"])

    def ui_filename(self):
        return "slits.ui"

    @property
    def caqtdm_ui_file(self):
        # Go up the class list until we find a class that is recognized
        for Cls in self.device.__class__.__mro__:
            try:
                return self.caqtdm_ui_filenames[Cls]
            except KeyError:
                continue
        # We didn't find any supported classes of slits
        msg = (
            "Could not find caQtDM filename for optic "
            f"{self.device.name} ({self.device.__class__})."
        )
        warnings.warn(msg)
        log.warning(msg)
        return ""

    def launch_caqtdm(self):
        # Sort out the prefix from the slit designator
        prefix = self.device.prefix.strip(":")
        pieces = prefix.split(":")
        # Build the macros for the caQtDM panels
        P = ":".join(pieces[:-1])
        SLIT = ":".join(pieces[-1:])
        H = self.device.h.prefix.split(":")[1]
        V = self.device.v.prefix.split(":")[1]
        caqtdm_macros = {
            "P": f"{P}:",
            # For 4-blade slits
            "SLIT": SLIT,
            "H": H,
            "V": V,
            # For rotary aperture slits
            "SLITS": SLIT,
        }
        # Add extra motors if applicable
        motors = {
            "HOR": "horizontal",
            "DIAG": "diagonal",
            "YAW": "yaw",
            "PITCH": "pitch",
        }
        for key, attr in motors.items():
            if not hasattr(self.device, attr):
                continue
            # Get the motor number from the device
            suffix = getattr(self.device, attr).prefix.split(":")[-1]
            caqtdm_macros[key] = suffix
        # Launch the caQtDM panel
        super().launch_caqtdm(macros=caqtdm_macros)


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



import os
from functools import partial

# This environmental variable needs to be set before importing haven
os.environ["HAVEN_CONFIG_FILES"] = f"{os.environ['BLUESKY_DIR']}/iconfig.toml"

import msgpack
import msgpack_numpy as mpn
from bluesky_kafka import MongoConsumer

import haven

bootstrap_servers = "localhost:9092"

# Determine the mongo DB URI from the databroker/intake configuration
catalog_name = haven.load_config()["database"]["databroker"]["catalog"]
catalog = haven.load_catalog(name=catalog_name)
host, port = catalog._resource_collection.database.client.address
mongo_uri = f"mongodb://{host}:{port}"

if mongo_uri is None:
    raise AttributeError("Environment variable BLUESKY_MONGO_URI " "must be set.")

kafka_deserializer = partial(msgpack.loads, object_hook=mpn.decode)
auto_offset_reset = "latest"
topics = ["25idc_queueserver", "25idd_queueserver"]

topic_database_map = {
    "25idc_queueserver": "25idc-bluesky",
    "25idd_queueserver": "25idd-bluesky",
}

# Create a MongoConsumer that will automatically listen to new beamline topics.
# The parameter metadata.max.age.ms determines how often the consumer will check for
# new topics. The default value is 5000ms.
mongo_consumer = MongoConsumer(
    mongo_uri,
    topic_database_map,
    tls=False,
    topics=topics,
    bootstrap_servers=bootstrap_servers,
    group_id="mongodb",
    consumer_config={"auto.offset.reset": auto_offset_reset},
    polling_duration=1.0,
    deserializer=kafka_deserializer,
)


mongo_consumer.start()



from unittest import mock

import pytest
from ophyd.sim import make_fake_device

from firefly.mirror import MirrorDisplay
from haven.instrument.mirrors import BendableHighHeatLoadMirror, HighHeatLoadMirror


@pytest.fixture()
def hhl_mirror(sim_registry):
    """A fake set of slits using the 4-blade setup."""
    FakeMirrors = make_fake_device(HighHeatLoadMirror)
    mirr = FakeMirrors(prefix="255ida:ORM1:", name="hhl_mirror", labels={"mirrors"})
    return mirr


@pytest.fixture()
def hhl_bendable_mirror(sim_registry):
    """A fake set of slits using the 4-blade setup."""
    FakeMirrors = make_fake_device(BendableHighHeatLoadMirror)
    mirr = FakeMirrors(prefix="255ida:ORM1:", name="hhl_mirror", labels={"mirrors"})
    return mirr


@pytest.fixture()
def display(qtbot, hhl_mirror):
    disp = MirrorDisplay(macros={"DEVICE": hhl_mirror.name})
    qtbot.addWidget(disp)
    return disp


def test_mirror_caqtdm(display):
    display._open_caqtdm_subprocess = mock.MagicMock()
    # Launch the caqtdm display
    display.launch_caqtdm()
    assert display._open_caqtdm_subprocess.called
    cmds = display._open_caqtdm_subprocess.call_args[0][0]
    # Check that the right macros are sent
    # /net/s25data/xorApps/epics/synApps_6_2/ioc/25ida/25idaApp/op/ui/HHLM_4.ui
    # macro: P=25ida:,MIR=ORM1:,Y=m1,ROLL=m2,LAT=lateral,CP=coarsePitch,,UPL=m3,DNL=m4
    macros = [cmds[i + 1] for i in range(len(cmds)) if cmds[i] == "-macro"][0]
    assert "P=255ida:" in macros
    assert "MIR=ORM1" in macros
    assert "Y=m1" in macros
    assert "ROLL=m2" in macros
    assert "LAT=lateral" in macros
    assert "CP=coarsePitch" in macros
    assert "UPL=m3" in macros
    assert "DNL=m4" in macros
    # Check that the right UI file is being used
    ui_file = cmds[-1]
    assert ui_file.split("/")[-1] == "HHLM_4.ui"


def test_bendable_mirror_caqtdm(hhl_bendable_mirror):
    mirror = hhl_bendable_mirror
    display = MirrorDisplay(macros={"DEVICE": mirror.name})
    display._open_caqtdm_subprocess = mock.MagicMock()
    # Launch the caqtdm display
    display.launch_caqtdm()
    assert display._open_caqtdm_subprocess.called
    cmds = display._open_caqtdm_subprocess.call_args[0][0]
    # Check that the right macros are sent
    # /net/s25data/xorApps/epics/synApps_6_2/ioc/25ida/25idaApp/op/ui/HHLM_4.ui
    # macro: P=25ida:,MIR=ORM1:,Y=m1,ROLL=m2,LAT=lateral,CP=coarsePitch,,UPL=m3,DNL=m4
    macros = [cmds[i + 1] for i in range(len(cmds)) if cmds[i] == "-macro"][0]
    assert "BEND=m5" in macros
    # Check that the right UI file is being used
    ui_file = cmds[-1]
    assert ui_file.split("/")[-1] == "HHLM_6.ui"


def test_bendable_mirror(hhl_bendable_mirror):
    mirror = hhl_bendable_mirror
    display = MirrorDisplay(macros={"DEVICE": mirror.name})
    # Check that the bender controls are unlocked
    assert display.ui.bender_embedded_display.isEnabled()

    # For a bendable HHL mirror
    # /net/s25data/xorApps/epics/synApps_6_2/ioc/25ida/25idaApp/op/ui/HHLM_6.ui, macro: P=25ida:,MIR=ORM2:,Y=m1,ROLL=m2,LAT=lateral,CP=coarsePitch,FP=PZT:m1,BEND=m5,UPL=m3,DNL=m4



from unittest import mock

import pytest

from firefly.slits import SlitsDisplay


@pytest.fixture()
def display(qtbot, blade_slits):
    disp = SlitsDisplay(macros={"DEVICE": blade_slits.name})
    qtbot.addWidget(disp)
    return disp


def test_blade_slit_caqtdm(display, blade_slits):
    display._open_caqtdm_subprocess = mock.MagicMock()
    # Launch the caqtdm display
    display.launch_caqtdm()
    assert display._open_caqtdm_subprocess.called
    cmds = display._open_caqtdm_subprocess.call_args[0][0]
    # Check that the right macros are sent
    macros = [cmds[i + 1] for i in range(len(cmds)) if cmds[i] == "-macro"][0]
    assert "P=255idc:" in macros
    assert "SLIT=KB_slits" in macros
    assert "H=KB_slitsH" in macros
    assert "V=KB_slitsV" in macros
    # Check that the right UI file is being used
    ui_file = cmds[-1]
    assert ui_file.split("/")[-1] == "4slitGraphic.ui"


def test_aperture_slit_caqtdm(display, aperture_slits):
    display._open_caqtdm_subprocess = mock.MagicMock()
    display.device = aperture_slits
    # Launch the caqtdm display
    display.launch_caqtdm()
    assert display._open_caqtdm_subprocess.called
    cmds = display._open_caqtdm_subprocess.call_args[0][0]
    # Check that the right macros are sent
    macros = [cmds[i + 1] for i in range(len(cmds)) if cmds[i] == "-macro"][0]
    assert "P=255ida:slits:" in macros
    assert "SLITS=US" in macros
    assert "HOR=m1" in macros
    assert "DIAG=m2" in macros
    assert "PITCH=m3" in macros
    assert "YAW=m4" in macros
    # Check that the right UI file is being used
    ui_file = cmds[-1]
    assert ui_file.split("/")[-1] == "maskApertureSlit.ui"
    # To-do: get the remaining macros for caQtDM
    # Full command should be:
    # /net/s25data/xorApps/epics/synApps_6_2/ioc/25ida/25idaApp/op/ui/maskApertureSlit.ui
    # macro: P=25ida:slits:,SLITS=US,HOR=m1,DIAG=m2,PITCH=m3,YAW=m4


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------



from collections import OrderedDict

import numpy as np
import pytest
from ophyd import EpicsMotor
from ophyd.flyers import FlyerInterface
from ophyd.sim import instantiate_fake_device
from ophyd.status import StatusBase

from haven.instrument.motor_flyer import MotorFlyer


@pytest.fixture()
def motor(sim_registry, mocker):
    Motor = type("Motor", (MotorFlyer, EpicsMotor), {})
    m = instantiate_fake_device(Motor, name="m1")
    mocker.patch.object(m, "move")
    m.user_setpoint._use_limits = False
    return m


def test_motor_flyer(motor):
    """Check that the haven motor implements the flyer interface."""
    assert motor is not None
    assert isinstance(motor, FlyerInterface)


def test_fly_params_forward(motor):
    """Test that the fly-scan parameters are correct when going from
    lower to higher positions.

    """
    # Set some example positions
    motor.motor_egu.set("micron").wait(timeout=3)
    motor.acceleration.set(0.5).wait(timeout=3)  # sec
    motor.flyer_start_position.set(10.0).wait(timeout=3)  # µm
    motor.flyer_end_position.set(20.0).wait(timeout=3)  # µm
    motor.flyer_num_points.set(101).wait(timeout=3)  # µm
    motor.flyer_dwell_time.set(1).wait(timeout=3)  # sec

    # Check that the fly-scan parameters were calculated correctly
    assert motor.flyer_slew_speed.get(use_monitor=False) == pytest.approx(0.1)  # µm/sec
    assert motor.flyer_taxi_start.get(use_monitor=False) == pytest.approx(9.9125)  # µm
    assert motor.flyer_taxi_end.get(use_monitor=False) == pytest.approx(20.0875)  # µm
    i = 10.0
    pixel = []
    while i <= 20.005:
        pixel.append(i)
        i = i + 0.1
    np.testing.assert_allclose(motor.pixel_positions, pixel)


def test_fly_params_reverse(motor):
    """Test that the fly-scan parameters are correct when going from
    higher to lower positions.

    """
    # Set some example positions
    motor.motor_egu.set("micron").wait(timeout=3)
    motor.acceleration.set(0.5).wait(timeout=3)  # sec
    motor.flyer_start_position.set(20.0).wait(timeout=3)  # µm
    motor.flyer_end_position.set(10.0).wait(timeout=3)  # µm
    motor.flyer_num_points.set(101).wait(timeout=3)  # µm
    motor.flyer_dwell_time.set(1).wait(timeout=3)  # sec

    # Check that the fly-scan parameters were calculated correctly
    assert motor.flyer_slew_speed.get(use_monitor=False) == pytest.approx(0.1)  # µm/sec
    assert motor.flyer_taxi_start.get(use_monitor=False) == pytest.approx(20.0875)  # µm
    assert motor.flyer_taxi_end.get(use_monitor=False) == pytest.approx(9.9125)  # µm
    i = 20.0
    pixel = []
    while i >= 9.995:
        pixel.append(i)
        i = i - 0.1
    np.testing.assert_allclose(motor.pixel_positions, pixel)


def test_kickoff(motor):
    motor.flyer_dwell_time.put(1.0)
    motor.flyer_taxi_start.put(1.5)
    # Start flying
    status = motor.kickoff()
    # Check status behavior matches flyer interface
    assert isinstance(status, StatusBase)
    status.wait(timeout=1)
    # Make sure the motor moved to its taxi position
    motor.move.assert_called_once_with(1.5, wait=True)


def test_complete(motor):
    # Set up fake flyer with mocked fly method
    assert motor.user_setpoint.get() == 0
    motor.flyer_taxi_end.put(10)
    # Complete flying
    status = motor.complete()
    # Check that the motor was moved
    assert isinstance(status, StatusBase)
    status.wait()
    motor.move.assert_called_once_with(10, wait=True)


def test_collect(motor):
    # Set up some fake positions from camonitors
    motor._fly_data = [
        # timestamp, position
        (1.125, 0.5),
        (2.125, 1.5),
        (3.125, 2.5),
        (4.125, 3.5),
        (5.125, 4.5),
        (6.125, 5.5),
        (7.125, 6.5),
        (8.125, 7.5),
        (9.125, 8.5),
        (10.125, 9.5),
        (11.125, 10.5),
    ]
    payload = list(motor.collect())
    # Confirm data have the right structure
    for datum, (timestamp, value) in zip(payload, motor._fly_data):
        assert datum["data"] == {
            "m1": value,
            "m1_user_setpoint": value,
        }
        assert datum["timestamps"]["m1"] == pytest.approx(timestamp, abs=0.3)
        assert datum["time"] == pytest.approx(timestamp, abs=0.3)


def test_predict(motor):
    # Set up some fake positions from camonitors
    motor._fly_data = [
        # timestamp, position
        (1.125, 0.5),
        (2.125, 1.5),
        (3.125, 2.5),
        (4.125, 3.5),
        (5.125, 4.5),
        (6.125, 5.5),
        (7.125, 6.5),
        (8.125, 7.5),
        (9.125, 8.5),
        (10.125, 9.5),
        (11.125, 10.5),
    ]
    # Prepare expected timestamp and position data
    expected_positions = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    motor.pixel_positions = np.asarray(expected_positions)
    expected_timestamps = [
        1.625,
        2.625,
        3.625,
        4.625,
        5.625,
        6.625,
        7.625,
        8.625,
        9.625,
        10.625,
    ]
    ts = [d[0] for d in motor._fly_data]
    vs = [d[1] for d in motor._fly_data]
    # Confirm data have the right structure
    for timestamp, expected_value in zip(expected_timestamps, expected_positions):
        datum = motor.predict(timestamp)
        assert datum["data"]["m1"] == pytest.approx(expected_value, abs=0.2)
        assert datum["data"]["m1_user_setpoint"] == expected_value
        assert datum["timestamps"]["m1"] == timestamp
        assert datum["time"] == timestamp


def test_describe_collect(aerotech_flyer):
    expected = {
        "positions": OrderedDict(
            [
                (
                    "aerotech_horiz",
                    {
                        "source": "SIM:aerotech_horiz",
                        "dtype": "integer",
                        "shape": [],
                        "precision": 3,
                    },
                ),
                (
                    "aerotech_horiz_user_setpoint",
                    {
                        "source": "SIM:aerotech_horiz_user_setpoint",
                        "dtype": "integer",
                        "shape": [],
                        "precision": 3,
                    },
                ),
            ]
        )
    }

    assert aerotech_flyer.describe_collect() == expected



import logging

from ophyd import Device
from ophyd import FormattedComponent as FCpt

from .._iconfig import load_config
from .device import make_device
from .motor import HavenMotor

__all__ = ["XYStage", "load_stages"]


log = logging.getLogger(__name__)


class XYStage(Device):
    """An XY stage with two motors operating in orthogonal directions.

    Vertical and horizontal are somewhat arbitrary, but are expected
    to align with the orientation of a camera monitoring the stage.

    Parameters
    ==========

    pv_vert
      The suffix to the PV for the vertical motor.
    pv_horiz
      The suffix to the PV for the horizontal motor.
    """

    vert = FCpt(HavenMotor, "{prefix}{pv_vert}", labels={"motors"})
    horiz = FCpt(HavenMotor, "{prefix}{pv_horiz}", labels={"motors"})

    def __init__(
        self,
        prefix: str,
        pv_vert: str,
        pv_horiz: str,
        labels={"stages"},
        *args,
        **kwargs,
    ):
        self.pv_vert = pv_vert
        self.pv_horiz = pv_horiz
        super().__init__(prefix, labels=labels, *args, **kwargs)


def load_stages(config=None):
    """Load the stages defined in the configuration files' ``[stage]``
    sections.

    """
    if config is None:
        config = load_config()
    devices = []
    for name, stage_data in config.get("stage", {}).items():
        devices.append(
            make_device(
                XYStage,
                name=name,
                prefix=stage_data["prefix"],
                pv_vert=stage_data["pv_vert"],
                pv_horiz=stage_data["pv_horiz"],
            )
        )
    return devices


# -----------------------------------------------------------------------------
# :author:    Mark Wolfman
# :email:     wolfman@anl.gov
# :copyright: Copyright © 2023, UChicago Argonne, LLC
#
# Distributed under the terms of the 3-Clause BSD License
#
# The full license is in the file LICENSE, distributed with this software.
#
# DISCLAIMER
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# -----------------------------------------------------------------------------
